# -*- coding: utf-8 -*-
r"""
Generic Grid View Adapter

**Grid View operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~GridViewAdapter.cell_to_display` | Static method for typecasting cell content to widget display value
    :meth:`~GridViewAdapter.display_to_cell` | Instance method for typecasting widget display value to cell content
    :meth:`~GridViewAdapter.compute_cells` | Compute object cells as a dictionary { coordinate pair : integer }
    :meth:`~GridViewAdapter.from_cells` | Create a new Sage object from a cells dictionary
    :meth:`~GridViewAdapter._validate` | Validate a new object
    :meth:`~GridViewAdapter.get_cell` | Get the object cell content
    :meth:`~GridViewAdapter.set_cell` | Set the object cell content
    :meth:`~GridViewAdapter.addable_cells` | List addable cells
    :meth:`~GridViewAdapter.removable_cells` | List removable cells
    :meth:`~GridViewAdapter.add_cell` | Add a cell at given position
    :meth:`~GridViewAdapter.remove_cell` | Remove a cell from given position
    :meth:`~GridViewAdapter.append_row` | Append a row
    :meth:`~GridViewAdapter.insert_row` | Insert a row at given index
    :meth:`~GridViewAdapter.remove_row` | Remove a row at given index
    :meth:`~GridViewAdapter.append_column` | Append a column
    :meth:`~GridViewAdapter.insert_column` | Insert a column at given index
    :meth:`~GridViewAdapter.remove_column` | Remove a column at given index

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
import traitlets, sage.all
from sage.all import SageObject
from sage.misc.abstract_method import abstract_method
from six import text_type

import __main__
def eval_in_main(s):
    """
    Evaluate the expression `s` in the global scope

    TESTS ::

        sage: from sage_widget_adapters.generic_grid_view_adapter import eval_in_main
        sage: from sage.combinat.tableau import Tableaux
        sage: eval_in_main("Tableaux")
        <class 'sage.combinat.tableau.Tableaux'>
    """
    try:
        return eval(s, sage.all.__dict__)
    except:
        return eval(s, __main__.__dict__)

class GridViewAdapter(object):
    r"""
    A generic grid view adapter.

    ATTRIBUTES::

        * ``objclass`` -- object class for this adapter
        * ``constructorname`` -- name of the constructor that builds a math object from a list
        * ``traitclass`` -- cells trait class
        * ``celltype`` -- cell content object type (to be defined in subclasses)
        * ``cellzero`` -- cell content zero (to be defined in subclasses)
        * ``addablecelltype`` -- addable cell content zero (to be defined in subclasses) -- by default = celltype
        * ``addablecellzero`` -- addable cell content zero (to be defined in subclasses) -- by default == cellzero
    """
    objclass = SageObject
    constructorname = None
    traitclass = traitlets.Instance
    constructorname = None
    addablecelltype = None
    addablecellzero = None

    @staticmethod
    def cell_to_display(cell_content, display_type=text_type):
        r"""
        From a cell value `cell_content`,
        return widget display value.

        TESTS ::

            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: from six import text_type
            sage: GridViewAdapter.cell_to_display(1, text_type)
            '1'
            sage: GridViewAdapter.cell_to_display(True, bool)
            True
        """
        if display_type == text_type:
            return str(cell_content)
        return cell_content

    def display_to_cell(self, display_value, display_type=text_type):
        r"""
        From an unicode string `s`,
        return matching cell value.

        TESTS ::

            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: from six import text_type
            sage: a.display_to_cell('1', text_type)
            Traceback (most recent call last):
            ...
            AttributeError: 'GridViewAdapter' object has no attribute 'celltype'
        """
        if display_value:
            return self.celltype(display_value)
        return self.cellzero

    @staticmethod
    @abstract_method
    def compute_cells(obj):
        r"""
        From an object `obj`,
        return a dictionary { coordinates pair : integer }
        """

    @classmethod
    def _validate(cls, obj, constructorname=''):
        r"""
        From an object `obj`,
        Try to build an object of type `cls`.

        TESTS ::

            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: assert issubclass(GridViewAdapter._validate(pi).__class__, SageObject)
            sage: from sage.matrix.constructor import matrix
            sage: assert issubclass(GridViewAdapter._validate(pi, constructorname='matrix').__class__, BaseException)
        """
        try:
            if constructorname:
                return eval_in_main(constructorname)(obj)
            if cls.constructorname:
                return eval_in_main(cls.constructorname)(obj)
            if issubclass(obj.__class__, cls.objclass):
                return obj
            return cls.objclass(obj)
        except Exception as e:
            return e

    @classmethod
    @abstract_method
    def from_cells(cls, cells={}):
        r"""
        From a dictionary { coordinates pair : integer },
        return a Sage object.
        """

    @staticmethod
    def get_cell(obj, pos):
        r"""
        From an object and a tuple `pos`,
        return the object cell value at position `pos`.

        TESTS ::

            sage: from sage.matrix.constructor import Matrix
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: m = Matrix(QQ, 3, 3, range(9))/2
            sage: GridViewAdapter.get_cell(m, (1,2))
            5/2
            sage: from sage.combinat.tableau import Tableau
            sage: t = Tableau([[1, 2, 5, 6], [3, 7], [4]])
            sage: GridViewAdapter.get_cell(t, (1,1))
            7
            sage: GridViewAdapter.get_cell(t, (1,6))
            Traceback (most recent call last):
            ...
            ValueError: Cell '(1, 6)' does not exist!
            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: st = SkewTableau([[None,1,2],[3,4,5],[6]])
            sage: GridViewAdapter.get_cell(st, (0,0))
            sage: GridViewAdapter.get_cell(st, (1,1))
            4
        """
        try:
            return obj[pos[0]][pos[1]]
        except:
            pass
        try:
            l = [list(x) for x in obj]
        except:
            raise NotImplementedError("Adapter class method 'get_cell(obj, pos)' is not implemented.")
        try:
            return l[pos[0]][pos[1]]
        except:
            raise ValueError("Cell '%s' does not exist!" % str(pos))

    def make_dirty(self, l, dirty={}):
        r"""
        Append 'dirty' values to list 'l'.
        Return a list with no empty values.

        TESTS ::

            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: from sage.combinat.tableau import Tableau
            sage: t = Tableau([[1, 2, 5, 6], [3, 7], [4]])
            sage: ga = GridViewAdapter()
            sage: ga.cellzero = 0
            sage: ga.make_dirty(t.to_list(), {(1,2):42})
            [[1, 2, 5, 6], [3, 7, 42], [4]]
            sage: ga.make_dirty(t.to_list(), {(2,0):0})
            [[1, 2, 5, 6], [3, 7]]
        """
        for p in dirty:
            if p[0] < len(l):
                if p[1] < len(l[p[0]]):
                    if dirty[p] == self.cellzero:
                        del l[p[0]][p[1]]
                    else:
                        l[p[0]][p[1]] = dirty[p]
                elif len(l[p[0]]) == p[1] and dirty[p] != self.cellzero:
                    l[p[0]].append(dirty[p])
            else:
                for i in range(p[0] - len(l)):
                    l.append([])
                l.append([dirty[p]])
        return [val for val in l if val]

    def set_cell(self, obj, pos, val, dirty={}, constructorname=''):
        r"""
        From a Sage object, a position (pair of coordinates) `pos` and a value `val`,
        return a new Sage object.
        with a modified cell at position `pos`.

        TESTS ::

            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: from sage.combinat.tableau import Tableau
            sage: t = Tableau([[1, 2, 5, 6], [3, 7], [4]])
            sage: ga = GridViewAdapter()
            sage: ga.set_cell(t, (1,1), 8, constructorname='Tableau')
            [[1, 2, 5, 6], [3, 8], [4]]
            sage: ga.cellzero = 0
            sage: ga.set_cell(t, (0,3), 6, {(0,3):5}, constructorname='StandardTableau')
            [[1, 2, 5, 6], [3, 7], [4]]
            sage: from sage.matrix.constructor import Matrix, matrix
            sage: m = Matrix(QQ, 3, 3, range(9))/2
            sage: ga.set_cell(m, (0,1), 2/3, constructorname='matrix')
            [  0 2/3   1]
            [3/2   2 5/2]
            [  3 7/2   4]
            sage: ga.set_cell(m, (4,2), 1/2, constructorname='matrix')
            Traceback (most recent call last):
            ...
            ValueError: Position '(4, 2)' does not exist
        """
        try:
            l = [list(x) for x in obj]
        except:
            raise NotImplementedError("Adapter method 'set_cell(obj, pos, val)' is not implemented.")
        l = self.make_dirty(l, dirty)
        try:
            l[pos[0]][pos[1]] = val
        except:
            raise ValueError("Position '%s' does not exist" % str(pos))
        return self._validate(l, constructorname)

    @staticmethod
    @abstract_method(optional = True)
    def addable_cells(obj):
        r"""
        For Sage object `obj`,
        list those cell positions where a user could want to add a cell,
        and get a still valid Sage object for this adapter.
        """

    @staticmethod
    @abstract_method(optional = True)
    def removable_cells(obj):
        r"""
        For Sage object `obj`,
        list those cell positions where a user could want to remove a cell,
        and get a still valid Sage object for this adapter.
        """

    @abstract_method(optional = True)
    def add_cell(self, obj, pos, val, dirty={}):
        r"""
        This method should try to add a cell to object `obj`
        at position `pos` and with value `val`.
        """

    @abstract_method(optional = True)
    def remove_cell(self, obj, pos, dirty={}):
        r"""
        This method should try to remove a cell from object `obj`
        at position `pos`.
        """

    @abstract_method(optional = True)
    def append_row(self, obj, r=None):
        r"""
        This method should try to append a row to object `obj`
        with values from list `r`.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.append_row(S, (1,2,3)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """

    @abstract_method(optional = True)
    def insert_row(self, obj, index, r=None):
        r"""
        This method should try to insert a row to object `obj`
        at index `index`, with values from list `r`.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.insert_row(S, 1, (1,2,3)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """

    def add_row(self, obj, index=None, r=None):
        r"""
        An alias for appending/inserting a row.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.add_row(S, 1, (1,2,3,4))
            Traceback (most recent call last):
            ...
            NotImplementedError: Method 'insert_row' is not implemented.
        """
        if index:
            try:
                return self.insert_row(obj, index, r)
            except:
                raise NotImplementedError("Method 'insert_row' is not implemented.")
        else:
            try:
                return self.append_row(obj, r)
            except:
                raise NotImplementedError("Method 'append_row' is not implemented.")

    @abstract_method(optional = True)
    def remove_row(self, obj, index=None):
        r"""
        This method should try to remove a row from object `obj`
        at index `index`.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.remove_row(S, 1) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """

    @abstract_method(optional = True)
    def append_column(self, obj, r=None):
        r"""
        This method should try to append a column to object `obj`
        with values from list `r`.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.append_column(S, (1,2,3,4)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """

    @abstract_method(optional = True)
    def insert_column(self, obj, index, r=None):
        r"""
        This method should try to insert a column to object `obj`
        at index `index`, with values from list `r`.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.insert_column(S, 1, (1,2,3,4)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """

    @abstract_method(optional = True)
    def remove_column(self, obj, index=None):
        r"""
        This method should try to remove a column from object `obj`
        at index `index`.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.remove_column(S, 2) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            TypeError: 'AbstractMethod' object is not callable
        """

    def add_column(self, obj, index=None, r=None):
        r"""
        An alias for appending/inserting a column.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
            sage: a = GridViewAdapter()
            sage: a.add_column(S, 1, (1,2,3))
            Traceback (most recent call last):
            ...
            NotImplementedError: Method 'insert_column' is not implemented.
        """
        if index:
            try:
                return self.insert_column(obj, index, r)
            except:
                raise NotImplementedError("Method 'insert_column' is not implemented.")
        else:
            try:
                return self.append_column(obj, r)
            except:
                raise NotImplementedError("Method 'append_column' is not implemented.")



# -*- coding: utf-8 -*-
r"""
Grid View Adapter for matrices

**Grid View matrix operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~MatrixGridViewAdapter.display_to_cell` | Instance method for typecasting widget display value to cell content
    :meth:`~MatrixGridViewAdapter.compute_cells` | Compute matrix cells as a dictionary { coordinate pair : label }
    :meth:`~MatrixGridViewAdapter.from_cells` | Create a new matrix from a cells dictionary
    :meth:`~MatrixGridViewAdapter.addable_cells` | List addable cells
    :meth:`~MatrixGridViewAdapter.removable_cells` | List removable cells
    :meth:`~MatrixGridViewAdapter.append_row` | Append a row
    :meth:`~MatrixGridViewAdapter.insert_row` | Insert a row at given index
    :meth:`~MatrixGridViewAdapter.remove_row` | Remove a row at given index
    :meth:`~MatrixGridViewAdapter.append_column` | Append a column
    :meth:`~MatrixGridViewAdapter.insert_column` | Insert a column at given index
    :meth:`~MatrixGridViewAdapter.remove_column` | Remove a column at given index

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
from sage.matrix.matrix2 import Matrix
from sage.matrix.constructor import matrix
from itertools import product
from sage.modules.free_module_element import vector
from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
from six import text_type

class MatrixGridViewAdapter(GridViewAdapter):
    r"""
    Grid view adapter for matrices.
    """
    objclass = Matrix
    constructorname = 'matrix'

    def __init__(self, obj):
        r"""
        Init an adapter object, set attributes `celltype` and `cellzero`.

        TESTS ::

            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: from sage.matrix.constructor import Matrix
            sage: m = Matrix(QQ, 3, 3, range(9))/2
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.celltype
            <type 'sage.rings.rational.Rational'>
            sage: ma.cellzero
            0
        """
        super(MatrixGridViewAdapter, self).__init__()
        self.ring = obj.base_ring()
        try:
            self.celltype = self.ring.element_class
        except:
            try:
                if hasattr(self.ring, 'an_element'):
                    self.celltype = self.ring.an_element().__class__
                elif hasattr(self.ring, 'random_element'):
                    self.celltype = self.ring.random_element().__class__
                else:
                    raise TypeError("Cannot determine matrix base ring elements class.")
            except:
                raise TypeError("Cannot determine matrix base ring elements class.")
        self.cellzero = self.ring.zero()

    def display_to_cell(self, display_value, display_type=text_type):
        r"""
        From a widget display value `display_value`,
        return matching cell value.

        TESTS ::

            sage: from sage.matrix.constructor import Matrix
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: m = Matrix(QQ, 3, 2, range(6))/2
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.display_to_cell('2/3')
            2/3
            sage: ma.display_to_cell('pi')
            Traceback (most recent call last):
            ...
            ValueError: Cannot cast display value pi to matrix cell
        """
        if display_value:
            try:
                return self.ring(display_value)
            except:
                try:
                    return self.celltype(display_value)
                except:
                    raise ValueError("Cannot cast display value %s to matrix cell" % (display_value))
        return self.cellzero

    @staticmethod
    def compute_cells(obj):
        r"""
        From a matrix `obj`,
        return a dictionary { coordinates pair : cell value (as a Sage object) }

        TESTS ::

            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: from sage.matrix.constructor import Matrix
            sage: m = Matrix(QQ, 3, 2, range(6))/2
            sage: MatrixGridViewAdapter.compute_cells(m)
            {(0, 0): 0, (0, 1): 1/2, (1, 0): 1, (1, 1): 3/2, (2, 0): 2, (2, 1): 5/2}
        """
        return {(i,j):obj[i][j] for (i,j) in product(range(obj.nrows()), range(len(obj[0])))}

    @classmethod
    def from_cells(cls, cells={}):
        r"""
        From a dictionary { coordinates pair : integer },
        return a matrix with corresponding cells.

        TESTS ::

            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: from sage.matrix.constructor import Matrix
            sage: MatrixGridViewAdapter.from_cells({(0, 0): 0, (0, 1): 1/2, (0, 2): 1, (1, 0): 3/2, (1, 1): 2, (1, 2): 5/2})
            [  0 1/2   1]
            [3/2   2 5/2]
        """
        nrows = max([pos[0]+1 for pos in cells])
        ncols = max([pos[1]+1 for pos in cells])
        return matrix([[cells[(i,j)] for j in range(ncols)] for i in range(nrows)])

    @staticmethod
    def addable_cells(obj):
        r"""
        No cell should be added in isolation
        except for vectors

        TESTS ::

            sage: from sage.matrix.constructor import Matrix
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: m = Matrix(QQ, 2, 3, range(6))/2
            sage: MatrixGridViewAdapter.addable_cells(m)
            []
        """
        if obj.nrows() == 1:
            return [(0, obj.ncols())]
        if obj.ncols() == 1:
            return [(obj.nrows(), 0)]
        return []

    @staticmethod
    def removable_cells(obj):
        r"""
        No cell should be removed in isolation
        except for vectors

        TESTS ::

            sage: from sage.matrix.constructor import Matrix
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: m = Matrix(QQ, 2, 3, range(6))/2
            sage: MatrixGridViewAdapter.removable_cells(m)
            []
        """
        if obj.nrows() == 1:
            return [(0, obj.ncols()-1)]
        if obj.ncols() == 1:
            return [(obj.nrows()-1, 0)]
        return []

    def remove_cell(self, obj, pos, dirty={}):
        r"""
        What to do if the user removes a value.
        (we replace the resulting blank with a zero)

        TESTS ::

            sage: from sage.matrix.constructor import Matrix
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: m = Matrix(QQ, 2, 3, range(6))/2
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.remove_cell(m, (1,0))
            [ 0 1/2   1]
            [ 0   2 5/2]
        """
        obj[pos] = self.cellzero
        return obj

    def append_row(self, obj, r=None):
        r"""
        Append a row to a matrix.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.append_row(m, (1,2,3))
            [ 1  7  1]
            [ 0  0  3]
            [ 0 -1  2]
            [ 1  0 -3]
            [ 1  2  3]
        """
        if not r:
            return obj.stack(vector([self.cellzero] * obj.ncols()))
        if len(r) > obj.ncols():
            print("Row is too long. Truncating")
            r = r[:obj.ncols()]
        elif len(r) < obj.ncols():
            r = list(r) + [self.cellzero] * (obj.ncols() - len(r))
        return obj.stack(vector([self.display_to_cell(x) for x in r]))

    def insert_row(self, obj, index, r=None):
        r"""
        Insert a row into a matrix.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.insert_row(m, 1, (1,2,3))
            [ 1  7  1]
            [ 1  2  3]
            [ 0  0  3]
            [ 0 -1  2]
            [ 1  0 -3]
        """
        if not r:
            r = [self.cellzero] * obj.ncols()
        else:
            if len(r) > obj.ncols():
                print("Row is too long. Truncating")
                r = r[:obj.ncols()]
            elif len(r) < obj.ncols():
                r = list(r) + [self.cellzero] * (obj.ncols() - len(r))
        top = obj.matrix_from_rows(range(index))
        bottom = obj.matrix_from_rows(range(index,obj.nrows()))
        return top.stack(vector([self.display_to_cell(x) for x in r])).stack(bottom)

    def remove_row(self, obj, index=None):
        r"""
        Remove a row from a matrix.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: m = S.matrix([0,1,2,3,4,5,6,7,8,9,10,11])
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.remove_row(m, 2)
            [ 0  1  2]
            [ 3  4  5]
            [ 9 10 11]
            sage: ma.remove_row(m)
            [0 1 2]
            [3 4 5]
            [6 7 8]
        """
        if index is None:
            index = obj.nrows() - 1
        return obj.delete_rows([index])

    def append_column(self, obj, c=None):
        r"""
        Append a column to a matrix.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.append_column(m, (1,1,1))
            [ 1  7  1  1]
            [ 0  0  3  1]
            [ 0 -1  2  1]
            [ 1  0 -3  0]
            sage: ma.append_column(m, (1,1,1,1,2,2))
            Column is too long. Truncating
            [ 1  7  1  1]
            [ 0  0  3  1]
            [ 0 -1  2  1]
            [ 1  0 -3  1]
        """
        if not c:
            return obj.augment(vector([self.cellzero]*obj.nrows()))
        if len(c) > obj.nrows():
            print("Column is too long. Truncating")
            c = c[:obj.nrows()]
        elif len(c) < obj.nrows():
            c = list(c) + [self.cellzero] * (obj.nrows() - len(c))
        return obj.augment(vector([self.display_to_cell(x) for x in c]))

    def insert_column(self, obj, index, c=None):
        r"""
        Insert a column into a matrix.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.insert_column(m, 1, (1,1,1))
            [ 1  1  7  1]
            [ 0  1  0  3]
            [ 0  1 -1  2]
            [ 1  0  0 -3]
            sage: ma.insert_column(m, 2, (1,1,1,2,2,2))
            Column is too long. Truncating
            [ 1  7  1  1]
            [ 0  0  1  3]
            [ 0 -1  1  2]
            [ 1  0  2 -3]
        """
        if not c:
            c = [self.cellzero] * obj.nrows()
        else:
            if len(c) > obj.nrows():
                print("Column is too long. Truncating")
                c = c[:obj.nrows()]
            elif len(c) < obj.nrows():
                c = list(c) + [self.cellzero] * (obj.nrows() - len(c))
        left = obj.matrix_from_columns(range(index))
        right = obj.matrix_from_columns(range(index,obj.ncols()))
        return left.augment(vector([self.display_to_cell(x) for x in c])).augment(right)

    def remove_column(self, obj, index=None):
        r"""
        Remove a column from a matrix.

        TESTS ::

            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
            sage: m = S.matrix([0,1,2,3,4,5,6,7,8,9,10,11])
            sage: ma = MatrixGridViewAdapter(m)
            sage: ma.remove_column(m, 1)
            [ 0  2]
            [ 3  5]
            [ 6  8]
            [ 9 11]
            sage: ma.remove_column(m)
            [ 0  1]
            [ 3  4]
            [ 6  7]
            [ 9 10]
        """
        if index is None:
            index = obj.ncols() - 1
        return obj.delete_columns([index])



# -*- coding: utf-8 -*-
r"""
An editable Grid View Widget for Sage Jupyter Notebook.

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
from .grid_view_editor import GridViewEditor, cdlink
from sage.graphs.generic_graph import GenericGraph
from ipywidgets import Layout, VBox, HBox, HTML, Text, ToggleButton, Button, ValueWidget
from singleton_widgets import *
from six import text_type

textcell_layout = Layout(width='3em', height='2em', margin='0', padding='0')
textcell_wider_layout = Layout(width='7em', height='3em', margin='0', padding='0')
buttoncell_layout = Layout(width='5em', height='4em', margin='0', padding='0')
buttoncell_smaller_layout = Layout(width='2em', height='2em', margin='0', padding='0')


class BaseTextCell(TextSingleton):
    r"""
    Abstract class for all text cells except blank.
    """
    displaytype = text_type

    def __init__(self, content, position, layout, **kws):
        super(BaseTextCell, self).__init__()
        self.value = content
        self.layout = layout
        self.continuous_update = False
        self.position = position
        self.description_tooltip = '' # avoid None
        self.add_class('gridcell')

class TextCell(BaseTextCell):
    r"""A regular text grid cell

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import TextCell
        sage: b = TextCell('my text', (1,2))
    """
    def __init__(self, content, position, layout=textcell_layout, **kws):
        super(TextCell, self).__init__(content, position, layout, **kws)

class StyledTextCell(TextCell):
    r"""A class for CSS-styled text grid cells.
    Not meant to be called directly.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import StyledTextCell
        sage: b = StyledTextCell("ok", (1,2))
        Traceback (most recent call last):
        ...
        TraitError: Element of the '_dom_classes' trait of a StyledTextCell instance must be a unicode string, but a value of None <type 'NoneType'> was specified.
    """
    disable = None
    css_class = None
    style = None
    def __init__(self, content, position, layout=textcell_layout, **kws):
        super(StyledTextCell, self).__init__(content, position, layout, **kws)
        self.add_class(self.css_class)
        if self.disable:
            self.disabled = True
        if self.style:
            apply_css(self.style)

def apply_css(css_line):
    try:
        ip = get_ipython()
        for base in ip.__class__.__mro__:
            """If we are in a notebook, we will find 'notebook' in those names"""
            if 'otebook' in base.__name__:
                ip.display_formatter.format(HTML("<style>%s</style>" % css_line))
                break
    except:
        pass # We are in the test environment

def styled_text_cell(disabled=False, style_name='', style=None):
    r"""A function to create CSS-styled cells.
    A regular text cell has a value and a position.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import styled_text_cell
        sage: styled_text_cell(disabled=True, style_name='mycssclass', style="")
        <class 'traitlets.traitlets.DisabledMycssclassTextCell'>
    """
    # FIXME passer la couleur en paramètre ? une chaîne CSS ?
    class_name = "{}TextCell".format(style_name.capitalize())
    if disabled:
        class_name = "Disabled" + class_name
    return type(class_name, (StyledTextCell,), {'disable': disabled, 'css_class': style_name, 'style': style})

class WiderTextCell(BaseTextCell):
    r"""A regular text grid cell

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import WiderTextCell
        sage: b = WiderTextCell('my text', (1,2))
    """
    def __init__(self, content, position, layout=textcell_wider_layout, **kws):
        super(WiderTextCell, self).__init__(content, position, layout, **kws)

class BlankCell(Text):
    r"""A blank placeholder cell

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import BlankCell
        sage: b = BlankCell()
    """
    displaytype = text_type

    def __init__(self, layout=textcell_layout, **kws):
        super(BlankCell, self).__init__()
        self.value = ''
        self.layout = layout
        self.disabled = True
        self.add_class('blankcell')

class AddableTextCell(BaseTextCell):
    r"""An addable placeholder for adding a cell to the widget

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import AddableTextCell
        sage: a = AddableTextCell((3,4))
    """
    def __init__(self, position, layout=textcell_layout):
        super(AddableTextCell, self).__init__('', position, layout=layout, continuous_update=False)
        self.add_class('addablecell')

class DisabledTextCell(BaseTextCell):
    r"""A disabled text grid cell

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import DisabledTextCell
        sage: b = DisabledTextCell('my text', (1,2))
    """
    def __init__(self, content, position, layout=textcell_layout, **kws):
        super(DisabledTextCell, self).__init__(content, position, layout=layout, **kws)
        self.disabled = True

class ButtonCell(ToggleButton):
    r"""A base class for button grid cells.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import ButtonCell
        sage: b = ButtonCell(True, (1,2))
    """
    displaytype = bool

    def __init__(self, content, position, layout=buttoncell_smaller_layout, **kws):
        super(ButtonCell, self).__init__()
        self.layout = layout
        self.value = content
        self.position = position
        self.add_class('gridbutton')
        self.set_tooltip()

    def set_tooltip(self, s=None):
        r"""From a position (i,j),
        we just want the string 'i,j'
        to use as a tooltip on buttons.

        TESTS ::

            sage: from sage_combinat_widgets.grid_view_widget import ButtonCell
            sage: b = ButtonCell(True, (42, 7))
            sage: b.set_tooltip()
            sage: str(b.tooltip)
            '42, 7'
            sage: b.set_tooltip("My new tooltip")
            sage: str(b.tooltip)
            'My new tooltip'
        """
        if s:
            self.tooltip = s
        else:
            self.tooltip = str(self.position)[1:-1]

class StyledButtonCell(ButtonCell):
    r"""A class for CSS-styled button grid cells.
    Not meant to be called directly.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import StyledButtonCell
        sage: b = StyledButtonCell(True, (1,2))
        Traceback (most recent call last):
        ...
        TraitError: Element of the '_dom_classes' trait of a StyledButtonCell instance must be a unicode string, but a value of None <type 'NoneType'> was specified.
    """
    disable = None
    css_class = None
    def __init__(self, content, position, layout=buttoncell_smaller_layout, **kws):
        super(StyledButtonCell, self).__init__(content, position, layout, **kws)
        self.add_class(self.css_class)
        if self.disable:
            self.disabled = True

def styled_button_cell(disabled=False, style_name=''):
    r"""A function to create CSS-styled buttons.
    A regular button has a value and a position.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import styled_button_cell
        sage: styled_button_cell(disabled=True, style_name='mycssclass')
        <class 'traitlets.traitlets.DisabledMycssclassButton'>
    """
    # FIXME passer la couleur en paramètre ? une chaîne CSS ?
    class_name = "{}Button".format(style_name.capitalize())
    if disabled:
        class_name = "Disabled" + class_name
    return type(class_name, (StyledButtonCell,), {'disable': disabled, 'css_class': style_name})

DisabledButtonCell = styled_button_cell(disabled=True)
r"""A disabled button cell.

TESTS ::

    sage: from sage_combinat_widgets.grid_view_widget import DisabledButtonCell
    sage: b = DisabledButtonCell(True, (1,2))
    sage: b.disabled
    True
"""

class AddableButtonCell(ButtonCell):
    r"""An addable placeholder for adding a button cell to the widget
    An addable button has a position.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import AddableButtonCell
        sage: a = AddableButtonCell((3,4))
    """
    def __init__(self, position, layout=buttoncell_smaller_layout, **kws):
        super(AddableButtonCell, self).__init__(False, position, layout, **kws)
        self.add_class('addablebutton')
        self.description = '+'
        self.tooltip = "Click to add a cell here"

class StyledPushButton(Button):
    r"""A class for CSS-styled push buttons.
    Not meant to be called directly.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import StyledPushButton
        sage: b = StyledPushButton()
    """
    disable = None
    css_class = None
    def __init__(self, content=None, position=None, layout=buttoncell_smaller_layout, description='', placeholder=None):
        super(StyledPushButton, self).__init__()
        self.layout=layout
        self.description=description
        if self.disable:
            self.disabled = True
        self.add_class('gridbutton')
        if self.css_class:
            self.add_class(self.css_class)

def styled_push_button(disabled=False, style_name=''):
    r"""A function to create CSS-styled push buttons.
    A push button stores neither value nor position.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import styled_push_button
        sage: styled_push_button(style_name='mycssclass').__name__
        'MycssclassPushButton'
    """
    class_name = "{}PushButton".format(style_name.capitalize())
    if disabled:
        class_name = "Disabled" + class_name
    return type(class_name, (StyledPushButton,), {'disable': disabled, 'css_class': style_name})

BlankButton = styled_push_button(disabled=True, style_name='blankbutton')
r"""A blank placeholder button.

TESTS ::

    sage: from sage_combinat_widgets.grid_view_widget import BlankButton
    sage: b = BlankButton()
    sage: b.__class__.__name__
    'DisabledBlankbuttonPushButton'
    sage: b.disabled
    True
    sage: assert 'blankbutton' in b._dom_classes
"""

def get_model_id(w):
    r"""
    For some reason, our widgets seem to lose their model_id
    This *hack* recovers it
    """
    for u in w.widgets:
        if w.widgets[u] == w:
            return u

class GridViewWidget(GridViewEditor, VBox, ValueWidget):
    r"""A widget for all grid-representable Sage objects
    """

    def __init__(self, obj, adapter=None, display_convention='en', cell_layout=None,
                 cell_widget_classes=[TextCell], cell_widget_class_index=lambda x:0,
                 blank_widget_class=BlankCell, addable_widget_class=AddableTextCell):
        r"""
        Grid View Widget initialization.

        INPUT:
            - ``cell_widget_classes``: a list of classes for building cell widgets
            - ``blank_widget_class``: a widget class for building blank cells
            - ``addable_widget_class``: a widget class for building blank cells

        TESTS ::

            sage: from sage_combinat_widgets.grid_view_widget import *
            sage: t = StandardTableaux(15).random_element()
            sage: w = GridViewWidget(t)
            sage: from sage.graphs.generators.families import AztecDiamondGraph
            sage: az = AztecDiamondGraph(4)
            sage: w = GridViewWidget(az, cell_widget_classes=[ButtonCell], blank_widget_class=BlankButton)

        Compatibility with `@interact`: the widget should be a
        :class:`ipywidgets.ValueWidget` and have a description field::

            sage: isinstance(w, ValueWidget)
            True
            sage: w.description
            "Grid view widget for Jupyter notebook with cell class '<class 'sage_combinat_widgets.grid_view_widget.ButtonCell'>', for object 'Subgraph of (2D Grid Graph for [8, 8])'"

        Basic compabitility test::

            sage: def f(x = w): return az.average_distance()
            sage: f = interact(f)
            <html>...</html>
        """
        GridViewEditor.__init__(self, obj, adapter)
        VBox.__init__(self)
        self._model_id = get_model_id(self)
        self.display_convention = display_convention
        self.description = "Grid view widget for Jupyter notebook with cell class '%s', for object '%s'" % (
            cell_widget_classes[0], obj)
        if not cell_layout:
            if issubclass(self.value.__class__, GenericGraph): # i.e. a graph
                cell_layout = buttoncell_smaller_layout
            else:
                cell_layout = textcell_layout
        self.cell_layout = cell_layout
        self.cell_widget_classes = cell_widget_classes
        self.cell_widget_class_index = cell_widget_class_index
        try:
            self.displaytype = cell_widget_classes[0].displaytype
        except:
            self.displaytype = None # Stateless cells
        self.cast = lambda x:self.adapter.display_to_cell(x, self.displaytype)
        self.blank_widget_class = blank_widget_class
        self.addable_widget_class = addable_widget_class
        self.draw()
        self.donottrack = False

    def to_cell(self, val):
        r"""
        From a widget cell value `val`,
        return a valid editor cell value.

        TESTS ::

            sage: from sage_combinat_widgets.grid_view_widget import GridViewWidget
            sage: t = StandardTableaux(5).random_element()
            sage: w = GridViewWidget(t)
            sage: w.to_cell('3')
            3
        """
        return self.cast(val)

    def add_links(self):
        r"""
        Link each individual widget cell
        to its corresponding trait in the editor

        TESTS ::

            sage: from sage.combinat.tableau import StandardTableaux
            sage: from sage_combinat_widgets.grid_view_widget import GridViewWidget
            sage: t = StandardTableaux(15).random_element()
            sage: w1 = GridViewWidget(t)
            sage: w2 = GridViewWidget(t, display_convention='fr')
            sage: len(w1.links)
            20
            sage: assert len(w1.links) == len(w2.links)
            sage: def test0(w): return (w.links[0].source[0].__class__, w.links[0].source[0].value, w.links[0].target[1])
            sage: def test10(w): return (w.links[10].source[0].__class__, w.links[10].source[0].value, w.links[10].target[1])
            sage: def test17(w): return (w.links[17].source[0].__class__, w.links[17].source[0].value, w.links[17].target[1])
            sage: assert test0(w1) == test0(w2)
            sage: assert test10(w1) == test10(w2)
            sage: assert test17(w1) == test17(w2)
            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: s = SkewTableau([[None, None, 1, 2], [None, 1], [4]])
            sage: w1 = GridViewWidget(s)
            sage: w2 = GridViewWidget(s, display_convention='fr')
            sage: len(w1.links)
            10
            sage: assert len(w1.links) == len(w2.links)
            sage: def test0(w): return (w.links[0].source[0].__class__, w.links[0].source[0].value, w.links[0].target[1])
            sage: def test4(w): return (w.links[4].source[0].__class__, w.links[4].source[0].value, w.links[4].target[1])
            sage: assert test0(w1) == test0(w2)
            sage: assert test4(w1) == test4(w2)
            sage: len(w2.links)
            10
            sage: w2.links[2].source[0].__class__
            <class 'sage_combinat_widgets.grid_view_widget.TextCell'>
            sage: w2.links[6].source[0].__class__
            <class 'sage_combinat_widgets.grid_view_widget.AddableTextCell'>
            sage: from traitlets import Bunch
            sage: w2.add_cell(Bunch({'name': 'add_0_4', 'old': 0, 'new': 3, 'owner': w2, 'type': 'change'}))
            sage: w2.value
            [[None, None, 1, 2, 3], [None, 1], [4]]
            sage: w2.links[2].source[0].__class__
            <class 'sage_combinat_widgets.grid_view_widget.TextCell'>
            sage: w2.links[7].source[0].__class__
            <class 'sage_combinat_widgets.grid_view_widget.AddableTextCell'>
        """
        for pos in self.cells.keys():
            traitname = 'cell_%d_%d' % (pos)
            child = self.get_child(pos)
            if child and hasattr(child, 'value') and traitname in self.traits():
                self.links.append(cdlink((child, 'value'), (self, traitname), self.cast))
        for pos in self.addable_cells():
            # A directional link to trait 'add_i_j'
            traitname = 'add_%d_%d' % (pos)
            child = self.get_child(pos)
            if child and hasattr(child, 'value') and traitname in self.traits():
                self.links.append(cdlink((child, 'value'), (self, traitname), self.cast))

    def draw(self, cell_widget_classes=None, cell_widget_class_index=None,
             addable_widget_class=None, blank_widget_class=None):
        r"""
        Add children to the GridWidget:
        - Sage object/grid editor cells
        - Blank cells for empty cells in a row
        - Addable cells if any
        Used classes can be passed as arguments
        to enable changing shapes, colors ..
        """
        self.donottrack = True # Prevent any interactivity while drawing the widget
        self.reset_links()
        self.compute_height()
        positions = sorted(list(self.cells.keys()))
        rows = [[(pos, self.cells[pos]) for pos in positions if pos[0]==i] \
                for i in range(self.height)]
        vbox_children = []
        addable_positions = self.addable_cells()
        addable_rows = []
        removable_positions = self.removable_cells()
        addable_rows = [(i,[pos for pos in addable_positions if pos[0]==i]) \
                        for i in set([t[0] for t in addable_positions])]
        if not cell_widget_classes:
            cell_widget_classes = self.cell_widget_classes
        if not cell_widget_class_index:
            cell_widget_class_index = self.cell_widget_class_index
        if not addable_widget_class:
            addable_widget_class = self.addable_widget_class
        if not blank_widget_class:
            blank_widget_class = self.blank_widget_class
        i, j = -1, -1 # initialization ; necessary for an empty grid
        for i in range(self.height):
            r = rows[i]
            if not r: # Empty row
                if (i,0) in addable_positions:
                    vbox_children.append(HBox((addable_widget_class((i,0), layout=self.cell_layout),)))
                else:
                    vbox_children.append(HBox((blank_widget_class(layout=self.cell_layout, disabled=True),)))
                continue
            j = 0
            hbox_children = []
            while j<=max([t[0][1] for t in rows[i]]):
                if (i,j) in positions:
                    cell_content = self.cells[(i,j)]
                    cell_widget_class = cell_widget_classes[cell_widget_class_index((i,j))]
                    cell_display = self.adapter.cell_to_display(cell_content, self.displaytype)
                    cell = cell_widget_class(cell_display,
                                             (i,j),
                                             layout=self.cell_layout,
                                             placeholder=cell_display)
                    if (i,j) in removable_positions:
                        if issubclass(cell_widget_class, ToggleButton):
                            cell.description = '-'
                            cell.disabled = False
                        else:
                            cell.add_class('removablecell')
                    hbox_children.append(cell)
                elif (i,j) in addable_positions:
                    # Inside the grid-represented object limits
                    hbox_children.append(addable_widget_class((i,j), layout=self.cell_layout))
                else:
                    hbox_children.append(blank_widget_class(layout=self.cell_layout))
                j+=1
                if j > max([t[0][1] for t in rows[i]]) and (i,j) in addable_positions:
                    # Outside of the grid-represented object limits
                    hbox_children.append(self.addable_widget_class((i,j), layout=self.cell_layout))
            vbox_children.append(HBox(hbox_children))
        for row in addable_rows:
            if row[0] >= self.height:
                hbox_children = []
                for c in range(max(row[1])[1]+1):
                    if (row[0],c) in row[1]:
                        hbox_children.append(self.addable_widget_class((i,j), layout=self.cell_layout))
                    else:
                        hbox_children.append(blank_widget_class(layout=self.cell_layout))
                vbox_children.append(HBox(hbox_children))
        if self.display_convention == 'fr':
            vbox_children.reverse()
        self.children = vbox_children
        self.add_links()
        self.donottrack = False

    def disallow_inside_focus(self):
        r"""
        Disallow focus for all cells except the first.
        """
        for r in self.children:
            for c in r.children:
                if hasattr(c, 'disallow_focus'):
                    c.disallow_focus()
        if hasattr(self.children[0].children[0], 'allow_focus'):
            self.children[0].children[0].allow_focus()

    def get_child(self, pos):
        r"""
        Get child widget corresponding to self.cells[pos]

        TESTS ::

            sage: from sage_combinat_widgets.grid_view_widget import GridViewWidget
            sage: t = StandardTableau([[1, 4, 7, 8, 9, 10, 11], [2, 5, 13], [3, 6], [12, 15], [14]])
            sage: w1 = GridViewWidget(t)
            sage: w1.get_child((1,2)).value
            u'13'
            sage: w2 = GridViewWidget(t, display_convention='fr')
            sage: w1.get_child((1,2)).value == w2.get_child((1,2)).value
            True
        """
        if self.display_convention == 'fr':
            return self.children[self.total_height - pos[0] - 1].children[pos[1]]
        return self.children[pos[0]].children[pos[1]]

    def set_dirty(self, pos, val, err=None):
        r"""
        Set cell #pos as dirty

        INPUT:

            - ``pos`` -- a tuple
            - ``val`` -- a(n incorrect) value for `pos`
            - ``err`` -- an exception

        TESTS ::

            sage: from sage_combinat_widgets import GridViewWidget
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: w = GridViewWidget(t)
            sage: from traitlets import Bunch
            sage: err = w.set_cell(Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': w, 'type': 'change'}))
            sage: w.set_dirty((0,2), 7, err)
            sage: w.dirty
            {(0, 2): 7}
            sage: w.dirty_errors[(0,2)]
            ValueError('the entries in each row of a semistandard tableau must be weakly increasing')
            sage: w.children[0].children[2]._dom_classes
            ('gridcell', 'dirty')
            sage: w.children[0].children[2]._tooltip
            'the entries in each row of a semistandard tableau must be weakly increasing'
        """
        super(GridViewWidget, self).set_dirty(pos, val, err)
        child = self.get_child(pos)
        child.add_class('dirty')
        if err:
            child.set_tooltip(self.dirty_info(pos))

    def unset_dirty(self, pos):
        r"""
        Set a cell no more 'dirty'.

        INPUT:

            - ``pos`` -- a tuple

        TESTS ::

            sage: from sage_combinat_widgets import GridViewWidget
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: w = GridViewWidget(t)
            sage: from traitlets import Bunch
            sage: err = w.set_cell(Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': e, 'type': 'change'}))
            sage: w.set_dirty((0,2), 7, err)
            sage: err = w.set_cell(Bunch({'name': 'cell_2_0', 'old': 4, 'new': 9, 'owner': e, 'type': 'change'}))
            sage: w.set_dirty((2,0), 9, err)
            sage: w.dirty
            {(0, 2): 7, (2, 0): 9}
            sage: w.unset_dirty((0,2))
            sage: w.dirty
            {(2, 0): 9}
            sage: w.children[0].children[2]._dom_classes
            ('gridcell',)
            sage: w.children[0].children[2]._tooltip
            ''
        """
        super(GridViewWidget, self).unset_dirty(pos)
        child = self.get_child(pos)
        child.remove_class('dirty')
        child.set_tooltip()

    def reset_dirty(self):
        r"""
        Reset all previously 'dirty' cells.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewWidget
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: w = GridViewWidget(t)
            sage: from traitlets import Bunch
            sage: err = w.set_cell(Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': e, 'type': 'change'}))
            sage: w.set_dirty((0,2), 7, err)
            sage: err = w.set_cell(Bunch({'name': 'cell_2_0', 'old': 4, 'new': 9, 'owner': e, 'type': 'change'}))
            sage: w.set_dirty((2,0), 9, err)
            sage: w.dirty
            {(0, 2): 7, (2, 0): 9}
            sage: w.children[2].children[0]._dom_classes
            ('gridcell', 'removablecell', 'dirty')
            sage: w.reset_dirty()
            sage: w.dirty
            {}
            sage: w.children[2].children[0]._dom_classes
            ('gridcell', 'removablecell')
            sage: w.children[2].children[0]._tooltip
            ''
        """
        for pos in self.dirty:
            child = self.get_child(pos)
            child.remove_class('dirty')
            child.set_tooltip()
        super(GridViewWidget, self).reset_dirty()

def PartitionGridViewWidget(obj, display_convention='en'):
    r"""
    A default widget for partitions.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_widget import PartitionGridViewWidget
        sage: sp = SkewPartition([[7, 4, 2, 1],[2, 1, 1]])
        sage: w = PartitionGridViewWidget(sp,  display_convention='fr')
        sage: len(w.links)
        17
    """
    return GridViewWidget(obj, cell_widget_classes=[ButtonCell],
                          addable_widget_class=AddableButtonCell, display_convention=display_convention)



# -*- coding: utf-8 -*-
r"""
Grid View Adapter for parallelogram polyominos

**Grid View parallelogram polyominos operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~ParallelogramPolyominoGridViewAdapter.compute_cells` | Compute parallelogram polyomino celss as a dictionary { coordinate pair : False }
    :meth:`~ParallelogramPolyominoGridViewAdapter.addable_cells` | List addable cells
    :meth:`~ParallelogramPolyominoGridViewAdapter.removable_cells` | List removable cells
    :meth:`~ParallelogramPolyominoGridViewAdapter.add_cell` | Add a cell
    :meth:`~ParallelogramPolyominoGridViewAdapter.remove_cell` | Remove a cell

AUTHORS ::

    Henri Derycke

"""
from sage.combinat.parallelogram_polyomino import ParallelogramPolyomino
from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter

class ParallelogramPolyominoGridViewAdapter(GridViewAdapter):
    r"""
    Grid view adapter for parallelogram polyominos.

    ATTRIBUTES::
        * ``objclass`` -- ParallelogramPolyomino
        * ``celltype`` -- bool
        * ``cellzero`` -- False
    """
    objclass = ParallelogramPolyomino
    celltype = bool
    cellzero = False

    @staticmethod
    def compute_cells(obj):
        r"""
        From a parallelogram polyomino,
        return a dictionary { coordinates pair : False }

        TESTS ::

            sage: from sage.combinat.parallelogram_polyomino import ParallelogramPolyomino
            sage: from sage_widget_adapters.combinat.parallelogram_polyomino_grid_view_adapter import ParallelogramPolyominoGridViewAdapter
            sage: pp = ParallelogramPolyomino([[0, 1, 1],[1, 1 ,0]])
            sage: ParallelogramPolyominoGridViewAdapter.compute_cells(pp)
            {(0, 0): True, (0, 1): True}
        """
        cells = {}
        lower_heights = obj.lower_heights()
        upper_heights = obj.upper_heights()
        for i in range(obj.width()):
            for j in range(upper_heights[i],lower_heights[i]):
                cells[j,i] = True
        return cells

    @staticmethod
    def addable_cells(obj):
        r"""
        List object addable cells

        TESTS ::

            sage: from sage.combinat.parallelogram_polyomino import ParallelogramPolyomino
            sage: from sage_widget_adapters.combinat.parallelogram_polyomino_grid_view_adapter import ParallelogramPolyominoGridViewAdapter
            sage: pp = ParallelogramPolyomino([[0, 1, 0, 1], [1, 1, 0, 0]])
            sage: ParallelogramPolyominoGridViewAdapter.addable_cells(pp)
            [(1, 0), (2, 1), (1, 2)]
        """
        cells = []

        upper_heights = obj.upper_heights()
        for i,c in enumerate(upper_heights[1:]):
            if c != upper_heights[i]:
                cells.append((c-1,i+1))

        lower_heights = obj.lower_heights()
        for i,c in enumerate(lower_heights[1:]):
            if c != lower_heights[i]:
                cells.append((lower_heights[i],i))

        height, width = obj.geometry()
        cells += [(height,width-1), (height-1,width)]

        return cells

    @staticmethod
    def removable_cells(obj):
        r"""
        List object removable cells

        TESTS ::

            sage: from sage.combinat.parallelogram_polyomino import ParallelogramPolyomino
            sage: from sage_widget_adapters.combinat.parallelogram_polyomino_grid_view_adapter import ParallelogramPolyominoGridViewAdapter
            sage: pp = ParallelogramPolyomino([[0, 0, 1, 1], [1, 1, 0, 0]])
            sage: ParallelogramPolyominoGridViewAdapter.removable_cells(pp)
            [(1, 0), (0, 1)]
        """
        heights = [(0,0)] + list(zip(obj.upper_heights(),obj.lower_heights()))
        heights.append((heights[-1][1],)*2)
        cells = []
        for i in range(1,len(heights)-1):
            x1,y1 = heights[i-1]
            x2,y2 = heights[i]
            x3,y3 = heights[i+1]
            if x2+1 < y2 and x2 != x3 and x2+1 < y1:
                cells.append((x2,i-1))
            if x2 < y2-1 and y1 != y2 and y2-1 > x3:
                cells.append((y2-1,i-1))

        if len(heights) > 3:
            x1,y1 = heights[-3]
            x2,y2 = heights[-2]
            if y1 < y2 or x2+1 == y2:
                cells.append((y2-1, len(heights)-3))

        return cells

    def add_cell(self, obj, pos, val=None, dirty={}):
        r"""
        Add cell

        TESTS ::

            sage: from sage.combinat.parallelogram_polyomino import ParallelogramPolyomino
            sage: from sage_widget_adapters.combinat.parallelogram_polyomino_grid_view_adapter import ParallelogramPolyominoGridViewAdapter
            sage: pp = ParallelogramPolyomino([[0, 1, 0, 1], [1, 1, 0, 0],])
            sage: ppa = ParallelogramPolyominoGridViewAdapter()
            sage: ppa.add_cell(pp, (1, 0))
            [[0, 0, 1, 1], [1, 1, 0, 0]]
            sage: ppa.add_cell(pp, (1, 1))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(1, 1)' is not addable.
        """
        if pos not in self.addable_cells(obj):
            raise ValueError("Cell position '%s' is not addable." % str(pos))

        heights = list(zip(obj.upper_heights(),obj.lower_heights()))

        upper_path = obj.upper_path()
        lower_path = obj.lower_path()

        height, width = obj.geometry()

        i,j = pos
        if i < height and j < width:
            index = i+j
            if heights[j][0] == i+1:
                upper_path[index:index+2] = [1,0]
            if heights[j][1] == i:
                lower_path[index:index+2] = [0,1]
        else:
            if i == height:
                lower_path[-1:] = [0,1]
                upper_path += [0]
            else:
                lower_path += [1]
                upper_path[-1:] = [1,0]

        return ParallelogramPolyomino([lower_path, upper_path])

    def remove_cell(self, obj, pos, dirty={}):
        r"""
        Remove cell

        TESTS ::

            sage: from sage.combinat.parallelogram_polyomino import ParallelogramPolyomino
            sage: from sage_widget_adapters.combinat.parallelogram_polyomino_grid_view_adapter import ParallelogramPolyominoGridViewAdapter
            sage: pp = ParallelogramPolyomino([[0, 0, 1, 1], [1, 1, 0, 0]])
            sage: ppa = ParallelogramPolyominoGridViewAdapter()
            sage: ppa.remove_cell(pp, (1, 0))
            [[0, 1, 0, 1], [1, 1, 0, 0]]
            sage: ppa.remove_cell(pp, (1, 1))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(1, 1)' is not removable.
        """
        if pos not in self.removable_cells(obj):
            raise ValueError("Cell position '%s' is not removable." % str(pos))

        heights = list(zip(obj.upper_heights(),obj.lower_heights()))
        upper_path = obj.upper_path()
        lower_path = obj.lower_path()
        i,j = pos
        index = i+j

        if len(heights) != j+1:
            if heights[j][0] == i:
                upper_path[index:index+2] = [0,1]
            if heights[j][1]-1 == i:
                lower_path[index:index+2] = [1,0]
        else:
            if heights[j][0] != i and heights[j][1]-1 == i:
                lower_path[index:index+2] = [1]
                upper_path.pop()
            elif heights[j][1]-1 == i:
                lower_path.pop()
                upper_path[index:index+2] = [0]
            else:
                upper_path[index:index+2] = [0,1]

        return ParallelogramPolyomino([lower_path, upper_path])




# -*- coding: utf-8 -*-
r"""
An editable Grid View Editor for Sage objects.

EXAMPLES ::

    sage: from sage_combinat_widgets import GridViewEditor
    sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
    sage: e = GridViewEditor(t)
    sage: e.addable_cells()
    [(0, 4), (1, 1), (3, 0)]
    sage: from sage.graphs.generators.basic import GridGraph
    sage: g = GridGraph((4,3))
    sage: e = GridViewEditor(g)
    sage: R = PolynomialRing(QQ, 9, 'x')
    sage: A = matrix(R, 3, 3, R.gens())
    sage: e = GridViewEditor(A)

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
import re, traitlets
from six import add_metaclass
from copy import copy
from sage.misc.bindable_class import BindableClass
from sage.all import SageObject
from sage.misc.abstract_method import AbstractMethod
MAX_LEN_HISTORY = 50

def extract_coordinates(s):
    r"""
    Extract a coordinate pair from a string with tokens.

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_editor import extract_coordinates
        sage: extract_coordinates('add_0_4')
        (0, 4)
    """
    patt = re.compile('_([0-9]+)_([0-9]+)')
    m = patt.search(s)
    if m:
        return tuple(int(i) for i in m.groups())

def get_adapter(obj):
    r"""
    Return an adapter object for Sage object `obj`.

    INPUT:

        - ``obj`` -- a Sage object

    OUTPUT: an adapter object

    TESTS ::

        sage: from sage_combinat_widgets.grid_view_editor import get_adapter
        sage: from sage.combinat.partition import Partition
        sage: p = Partition([3,2,1,1])
        sage: pa = get_adapter(p)
        sage: pa.cellzero
        False
        sage: from sage.combinat.tableau import StandardTableaux
        sage: t = StandardTableaux(7).random_element()
        sage: ta = get_adapter(t)
        sage: ta.cellzero
        0
    """
    from sage.combinat.partition import Partition
    if issubclass(obj.__class__, Partition):
        from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
        return PartitionGridViewAdapter()
    from sage.combinat.skew_partition import SkewPartition
    if issubclass(obj.__class__, SkewPartition):
        from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
        return SkewPartitionGridViewAdapter()
    from sage.combinat.tableau import Tableau
    if issubclass(obj.__class__, Tableau):
        from sage.combinat.tableau import SemistandardTableau, StandardTableau
        if issubclass(obj.__class__, StandardTableau):
            from sage_widget_adapters.combinat.tableau_grid_view_adapter import StandardTableauGridViewAdapter
            return StandardTableauGridViewAdapter()
        if issubclass(obj.__class__, SemistandardTableau):
            from sage_widget_adapters.combinat.tableau_grid_view_adapter import SemistandardTableauGridViewAdapter
            return SemistandardTableauGridViewAdapter()
        from sage_widget_adapters.combinat.tableau_grid_view_adapter import TableauGridViewAdapter
        return TableauGridViewAdapter()
    from sage.combinat.skew_tableau import SkewTableau
    if issubclass(obj.__class__, SkewTableau):
        from sage_widget_adapters.combinat.skew_tableau_grid_view_adapter import SkewTableauGridViewAdapter
        return SkewTableauGridViewAdapter()
    try:
        from sage.combinat.parallelogram_polyomino import ParallelogramPolyomino
        if issubclass(obj.__class__, ParallelogramPolyomino):
            from sage_widget_adapters.combinat.parallelogram_polyomino_grid_view_adapter import ParallelogramPolyominoGridViewAdapter
            return ParallelogramPolyominoGridViewAdapter()
    except:
        pass # support for parallelogram polyomino appears only in Sage 8.9
    from sage.matrix.matrix2 import Matrix
    if issubclass(obj.__class__, Matrix):
        from sage_widget_adapters.matrix.matrix_grid_view_adapter import MatrixGridViewAdapter
        return MatrixGridViewAdapter(obj)
    from sage.graphs.graph import Graph
    if issubclass(obj.__class__, Graph):
        from sage_widget_adapters.graphs.graph_grid_view_adapter import GraphGridViewAdapter
        return GraphGridViewAdapter()

class cdlink(traitlets.dlink):
    def __repr__(self):
        return "A typecasting directional link from source=(%s, %s) to target='%s'" % (
            self.source[0].__class__, self.source[0].value, self.target[1])

import sage.misc.classcall_metaclass
class MetaHasTraitsClasscallMetaclass(traitlets.MetaHasTraits, sage.misc.classcall_metaclass.ClasscallMetaclass):
    pass
@add_metaclass(MetaHasTraitsClasscallMetaclass)
class BindableClassWithMeta(BindableClass):
    pass
class BindableEditorClass(traitlets.HasTraits, BindableClassWithMeta):
    pass

class GridViewEditor(BindableEditorClass):
    r"""Base Editor for grid-representable Sage Objects

    Composed of cells. No decision made here about cell representation.
    The cell trait objects will store values
    that are refered to through object cells as a dictionary
    with coordinates (row_number, cell_number_in_row) as keys
    """
    value = traitlets.Any()

    def __init__(self, obj, adapter=None):
        r"""
        Initialize editor.

        INPUT:

            - ``obj`` -- a Sage object
            - ``adapter`` -- an adapter object (optional)

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: from sage.graphs.generic_graph import GenericGraph
            sage: e = GridViewEditor(graphs.AztecDiamondGraph(2))
            sage: e.cells
            {(0, 1): None, (0, 2): None, (1, 0): None, (1, 1): None, (1, 2): None, (1, 3): None, (2, 0): None,
            (2, 1): None, (2, 2): None, (2, 3): None, (3, 1): None, (3, 2): None}
            sage: e = GridViewEditor(StandardTableaux(5).random_element())
            sage: e.cells
            {(0, 0): 1, (1, 0): 2, (2, 0): 3, (3, 0): 4, (4, 0): 5}
            sage: e = GridViewEditor(vector((1,2,3)))
            Traceback (most recent call last):
            ...
            TypeError: Cannot find an Adapter for this object (<type 'sage.modules.vector_integer_dense.Vector_integer_dense'>)
            sage: e = GridViewEditor(x^5)
            Traceback (most recent call last):
            ...
            TypeError: Cannot find an Adapter for this object (<type 'sage.symbolic.expression.Expression'>)
        """
        self.donottrack = True
        self.dirty = {}
        super(GridViewEditor, self).__init__()
        self.value = obj
        self._initval = obj
        self._history = []
        self.dirty_errors = {}
        if adapter:
            self.adapter = adapter
        else:
            self.adapter = get_adapter(obj)
        if not self.adapter:
            raise TypeError("Cannot find an Adapter for this object (%s)" % obj.__class__)
        if not hasattr(self.adapter, 'compute_cells') or not callable(self.adapter.compute_cells):
            raise NotImplementedError("Method `compute_cells` is required!")
        self.compute()
        self.links = []

    def to_cell(self, val):
        r"""
        From a widget cell value `val`,
        return a valid editor cell value.
        Will be overloaded in widget code.
        """
        return val

    def validate(self, new_obj, obj_class=None):
        r"""
        Validate object type.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: obj = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(obj)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: new_valid_obj = Tableau([[1, 2, 3, 6], [4], [5]])
            sage: e.validate(new_valid_obj, obj.__class__)
            True
            sage: new_invalid_obj = Partition([3,3,1])
            sage: issubclass(new_invalid_obj.__class__, obj.__class__)
            False
            sage: e.validate(new_invalid_obj, obj.__class__)
            False
            sage: new_invalid_obj = 42
            sage: e.validate(new_invalid_obj)
            False
        """
        if obj_class:
            return issubclass(new_obj.__class__, obj_class)
        return issubclass(new_obj.__class__, SageObject) and hasattr(new_obj, 'cells')

    def modified_add_traits(self, **traits):
        r"""
        Dynamically add trait attributes to the HasTraits instance.
        Modified code according to Ryan Morshead's pull request
        Cf https://github.com/ipython/traitlets/pull/501
        """
        cls = self.__class__
        attrs = {"__module__": cls.__module__}
        if hasattr(cls, "__qualname__"):
          # __qualname__ introduced in Python 3.3 (see PEP 3155)
          attrs["__qualname__"] = cls.__qualname__
        attrs.update(traits)
        self.__class__ = type(cls.__name__, (cls,), attrs)
        for trait in traits.values():
            trait.instance_init(self)

    def compute(self, obj=None):
        r"""We have an object value
        but we want to compute cells
        as a dictionary (row_number, cell_number_in_row) -> trait
        """
        if not obj:
            obj = self.value
        if obj is None:
            return
        self.cells = self.adapter.compute_cells(obj)
        celltype = self.adapter.celltype
        cellzero = self.adapter.cellzero
        addablecelltype = self.adapter.addablecelltype or celltype
        addablecellzero = self.adapter.addablecellzero or cellzero
        traitclass = self.adapter.traitclass
        traits_to_add = {}
        for pos in self.addable_cells():
            # Empty traits for addable cells
            emptytraitname = 'add_%d_%d' % pos
            try:
                emptytrait = traitclass(addablecellzero)
            except:
                try:
                    emptytrait = traitclass(addablecelltype)
                except:
                    raise TypeError("Cannot init the trait (traitclass=%s, celltype=%s, default_value=%s)" % (
                        traitclass, addablecelltype, addablecellzero))
            emptytrait.name = emptytraitname
            traits_to_add[emptytraitname] = emptytrait
        for pos, val in self.cells.items():
            traitname = 'cell_%d_%d' % pos
            traitvalue = val
            if traitname in self._trait_values:
                self._trait_values[traitname] = traitvalue
            else:
                try:
                    trait = traitclass(traitvalue)
                except:
                    try:
                        trait = traitclass(celltype)
                        trait.value = traitvalue
                    except:
                        raise TypeError("Cannot init the trait (traitclass=%s, celltype=%s, default_value=%s)" % (
                            traitclass, celltype, cellzero))
                trait.name = traitname
                traits_to_add[traitname] = trait
        self.traitclass = traitclass
        self.modified_add_traits(**traits_to_add)

    def compute_height(self):
        r"""
        Compute grid height, addable cells included.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: from sage.combinat.partition import Partition
            sage: e = GridViewEditor(Partition([3,3,2,1]))
            sage: e.compute_height()
            sage: e.height
            4
            sage: from sage.graphs.generators.families import AztecDiamondGraph
            sage: e = GridViewEditor(AztecDiamondGraph(2))
            sage: e.compute_height()
            sage: e.height
            4
        """
        if not hasattr(self, 'cells'):
            self.compute()
        if self.cells:
            maxpos = max(pos[0] for pos in self.cells)
        else:
            maxpos = -1
        self.height = maxpos + 1 # Number of rows in self.value
        for pos in self.addable_cells():
            if pos[0] > maxpos:
                maxpos = pos[0]
        self.total_height =  maxpos + 1 # Graphical height

    def reset_links(self):
        r"""
        Reset all potentially existing links
        between widget cells and corresponding traits.
        """
        for lnk in self.links:
            lnk.unlink()
        self.links = []

    def draw(self, cast=None):
        r"""
        Build the visual representation
        and cdlink objects -- with cast function `cast`.
        """
        pass

    def get_value(self):
        r"""
        Return editor value.

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_combinat_widgets import GridViewEditor
            sage: e = GridViewEditor(Tableau([[1, 2, 5, 6], [3], [4]]))
            sage: e.get_value()
            [[1, 2, 5, 6], [3], [4]]
        """
        return self.value

    def set_value(self, obj):
        r"""
        Check compatibility, then set editor value.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: new_valid_obj = Tableau([[1, 2, 7, 6], [3], [4]])
            sage: e.set_value(new_valid_obj)
            sage: e.value
            [[1, 2, 7, 6], [3], [4]]
            sage: new_invalid_obj = 42
            sage: e.set_value(new_invalid_obj)
            Traceback (most recent call last):
            ...
            ValueError: Object 42 is not compatible. A tableau must be a list of iterables.
        """
        self.reset_dirty()
        res = self.adapter._validate(obj)
        if issubclass(res.__class__, BaseException):
            raise ValueError("Object %s is not compatible. %s" % (obj, res))
        self.value = obj # Will call the observer, but only if value has changed

    def push_history(self, obj):
        r"""
        Push an object to editor history.
        Ensure that history does not become too long.

        INPUT:

            - ``obj`` -- an object (the old one)

        TESTS::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e._history
            []
            sage: e.push_history(t)
            sage: e._history
            [[[1, 2, 5, 6], [3], [4]]]

        """
        self._history.append(obj)
        if len(self._history) > MAX_LEN_HISTORY:
            self._history = self._history[1:]

    @traitlets.observe('value')
    def value_changed(self, change):
        r"""
        What to do when the value has been changed.

        INPUT:

            - ``change`` -- a change Bunch

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: new_t = Tableau([[1, 2, 7, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: e._history
            []
            sage: from traitlets import Bunch
            sage: e.value_changed(Bunch({'name': 'value', 'old': t, 'new': new_t, 'owner': e, 'type': 'change'}))
            sage: e._history
            [[[1, 2, 5, 6], [3], [4]]]
        """
        self.reset_dirty()
        if self.donottrack:
            return
        old_val = change.old
        new_val = change.new
        actually_changed = (id(new_val) != id(old_val))
        if actually_changed:
            self.push_history(old_val)
            self.compute()
            self.draw()

    def pop_value(self):
        r"""
        """
        if not self._history:
            print("No more history!")
            return
        prev = self._history.pop()
        self.donottrack = True
        self.value = prev
        self.compute()
        self.draw()
        self.donottrack = False

    def get_cells(self):
        r"""
        Return grid editor cells.
        """
        return self.cells

    def set_value_from_cells(self, obj_class=None, cells={}):
        r"""We have an object value, but we want to change it according to cells
        Yet we want to keep the same class (or warn if that's impossible)

        INPUT:

            -  ``obj_class`` -- an object class (by default: self.value.__class__)
            -  ``cells`` -- a dictionary (i,j)->val
        """
        if not obj_class and self.value:
            obj_class = self.value.__class__
        if not obj_class:
            return
        if hasattr(self.adapter, 'from_cells'):
            try:
                obj = self.adapter.from_cells(cells)
            except:
                raise ValueError("Could not make an object of class '%s' from given cells" % str(obj_class))
        elif hasattr(obj_class, 'cells') or hasattr(obj_class, 'rows'): # e.g. a tableau / matrix / vector
            positions = sorted(list(cells.keys()))
            for cl in obj_class.__mro__:
                try:
                    obj = cl([[cells[pos] for pos in positions if pos[0]==i] for i in set([t[0] for t in positions])])
                except:
                    print("These cells cannot be turned into a %s" % cl)
        else:
            raise TypeError("Unable to cast the given cells into a grid-like object.")
        if not self.validate(obj, obj_class):
            raise ValueError("Could not make a compatible ('%s')  object from given cells" % str(obj_class))
        self.donottrack = True
        self.set_value(obj)
        self.donottrack = False

    def set_dirty(self, pos, val, err=None):
        r"""
        Set a cell 'dirty'.

        INPUT:

            - ``pos`` -- a tuple
            - ``val`` -- a(n incorrect) value for `pos`
            - ``err`` -- an exception

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: from traitlets import Bunch
            sage: err = e.set_cell(Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': e, 'type': 'change'}))
            sage: e.set_dirty((0,2), 7, err)
            sage: e.dirty
            {(0, 2): 7}
            sage: e.dirty_errors[(0,2)]
            ValueError('the entries in each row of a semistandard tableau must be weakly increasing')
        """
        self.dirty[pos] = val
        if err:
            self.dirty_errors[pos] = err

    def unset_dirty(self, pos):
        r"""
        Set a cell no more 'dirty'.

        INPUT:

            - ``pos`` -- a tuple

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: from traitlets import Bunch
            sage: err = e.set_cell(Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': e, 'type': 'change'}))
            sage: e.set_dirty((0,2), 7, err)
            sage: err = e.set_cell(Bunch({'name': 'cell_2_0', 'old': 4, 'new': 9, 'owner': e, 'type': 'change'}))
            sage: e.set_dirty((2,0), 9, err)
            sage: e.dirty
            {(0, 2): 7, (2, 0): 9}
            sage: e.unset_dirty((0,2))
            sage: e.dirty
            {(2, 0): 9}
        """
        del self.dirty[pos]
        del self.dirty_errors[pos]

    def reset_dirty(self):
        r"""
        Reset all previously 'dirty' cells.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: from traitlets import Bunch
            sage: err = e.set_cell(Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': e, 'type': 'change'}))
            sage: e.set_dirty((0,2), 7, err)
            sage: err = e.set_cell(Bunch({'name': 'cell_2_0', 'old': 4, 'new': 9, 'owner': e, 'type': 'change'}))
            sage: e.set_dirty((2,0), 9, err)
            sage: e.dirty
            {(0, 2): 7, (2, 0): 9}
            sage: e.reset_dirty()
            sage: e.dirty
            {}
        """
        if not self.dirty: # Prevent any interactive loops
            return
        self.dirty = {}
        self.dirty_errors = {}

    def dirty_info(self, pos):
        r"""
        Get error details from a 'dirty' cell.

        INPUT:

            - ``pos`` -- a tuple

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: from traitlets import Bunch
            sage: err = e.set_cell(Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': e, 'type': 'change'}))
            sage: e.set_dirty((0,2), 7, err)
            sage: err = e.set_cell(Bunch({'name': 'cell_2_0', 'old': 4, 'new': 9, 'owner': e, 'type': 'change'}))
            sage: e.set_dirty((2,0), 9, err)
            sage: e.dirty_info((0, 2))
            'the entries in each row of a semistandard tableau must be weakly increasing'
        """
        if pos in self.dirty_errors:
            return str(self.dirty_errors[pos])
        return ''

    @traitlets.observe(traitlets.All)
    def set_cell(self, change):
        r"""
        What to do when a cell value has been changed.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False # This class is not meant to work by itself without a widget.
            sage: from traitlets import Bunch
            sage: change = Bunch({'name': 'cell_0_2', 'old': 5, 'new': 7, 'owner': e, 'type': 'change'})
            sage: e.set_cell(change)
            sage: e.value
            [[1, 2, 7, 6], [3], [4]]
        """
        if self.donottrack:
            return
        if not change.name.startswith('cell_'):
            return
        if change.old == traitlets.Undefined: # Do nothing on a newly added cell
            return
        if change.new == change.old or not change.new:
            return
        pos = extract_coordinates(change.name)
        val = change.new
        result = self.adapter.set_cell(self.value, pos, val, dirty=self.dirty)
        if issubclass(result.__class__, BaseException): # Setting cell was impossible
            if val == self.cells[pos] and self.dirty.keys() == [pos]: # Rollback
                self.reset_dirty()
            else: # Add an entry in self.dirty dictionary
                self.set_dirty(pos, val, result)
            return
        # Success
        self.set_value(result)

    def addable_cells(self):
        r"""
        List addable cells for editor value
        """
        if not hasattr(self.adapter, 'addable_cells') or not callable(self.adapter.addable_cells):
            return [] # Optional method
        return self.adapter.addable_cells(self.value)

    def removable_cells(self):
        r"""
        List removable cells for editor value.
        """
        if not hasattr(self.adapter, 'removable_cells') or not callable(self.adapter.removable_cells):
            return [] # Optional method
        return self.adapter.removable_cells(self.value)

    @traitlets.observe(traitlets.All)
    def add_cell(self, change):
        r"""
        Add a cell to the widget.

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False
            sage: from traitlets import Bunch
            sage: change = Bunch({'name': 'add_1_1', 'old': 0, 'new': 8, 'owner': e, 'type': 'change'})
            sage: e.add_cell(change)
            sage: e.value
            [[1, 2, 5, 6], [3, 8], [4]]
            sage: t = StandardTableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False
            sage: e.add_cell(change)
            sage: e.value
            [[1, 2, 5, 6], [3], [4]]
            sage: e.dirty
            {(1, 1): 8}
            sage: e = GridViewEditor(SkewTableau([[None, None, 1, 2], [None, 1], [4]]))
            sage: e.donottrack = False
            sage: e.add_cell(Bunch({'name': 'add_0_4', 'old': 0, 'new': 3, 'owner': e, 'type': 'change'}))
            sage: e.value
            [[None, None, 1, 2, 3], [None, 1], [4]]
            sage: e.add_cell(Bunch({'name': 'add_1_0', 'old': 0, 'new': 1, 'owner': e, 'type': 'change'}))
            sage: e.value
            [[None, None, 1, 2, 3], [1, 1], [4]]
        """
        if self.donottrack:
            return
        if not change.name.startswith('add_') \
           or self.to_cell(change.new) == self.adapter.cellzero:
            return
        if not hasattr(self.adapter, 'add_cell'):
            raise TypeError("Cannot add cell to this object.")
        if self.adapter.add_cell.__func__.__class__ is AbstractMethod:
            return # Method not implemented
        if hasattr(self.adapter.remove_cell, '_optional') and self.adapter.remove_cell._optional: # Not implemented
            raise Exception("Adding cells is not implemented for this object.")
        val = change.new
        if val is True: # if it's a button, reverse button toggling
            val = False
        pos = extract_coordinates(change.name)
        if pos in self.dirty:
            self.dirty[pos] = val # edit the value before sending to the adapter
        obj = copy(self.value)
        result = self.adapter.add_cell(obj, pos, val, dirty=self.dirty)
        if issubclass(result.__class__, BaseException): # Adding cell was impossible
            if pos in self.cells and val == self.cells[pos] and self.dirty.keys() == [pos]: # Rollback
                self.reset_dirty()
            else: # Keep temporary addition for later
                self.set_dirty(pos, val, result)
            return
        self.set_value(result)

    @traitlets.observe(traitlets.All)
    def remove_cell(self, change):
        r"""
        What to do when a cell has been removed.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: t = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: e = GridViewEditor(t)
            sage: e.donottrack = False
            sage: from traitlets import Bunch
            sage: e.remove_cell(Bunch({'name': 'cell_0_3', 'old': 6, 'new': 0, 'owner': e, 'type': 'change'}))
            sage: e.value
            [[1, 2, 5], [3], [4]]
            sage: e.remove_cell(Bunch({'name': 'cell_2_0', 'old': 4, 'new': 0, 'owner': e, 'type': 'change'}))
            sage: e.value
            [[1, 2, 5], [3]]
            sage: e = GridViewEditor(SkewTableau([[None, None, 1, 2, 3], [None, 1], [4]]))
            sage: e.donottrack = False
            sage: e.remove_cell(Bunch({'name': 'cell_0_4', 'old': 3, 'new': 0, 'owner': e, 'type': 'change'}))
            sage: e.value
            [[None, None, 1, 2], [None, 1], [4]]
            sage: p = Partition([1])
            sage: e = GridViewEditor(p)
            sage: e.donottrack = False
            sage: e.remove_cell(Bunch({'name': 'cell_0_0', 'old': False, 'new': True, 'owner': e, 'type': 'change'}))
            sage: e.value
            []
        """
        # Do nothing at widget donottrack and do not track widget value
        if self.donottrack or change.name == 'value':
            return
        val = change.new
        if val is True: # if it's a button, reverse button toggling
            val = False
        pos = extract_coordinates(change.name)
        # Dirty _addable_ cells can be removed
        if pos in self.dirty and change.name.startswith('add_') and self.to_cell(val) == self.adapter.cellzero:
            self.unset_dirty(pos)
            return
        # Don't remove non empty cells
        # Don't remove addable cells ; this test will do something only if adapter addablecellzero is specified (therefore not eq to cellzero)
        if not change.name.startswith('cell_') or self.to_cell(val) != self.adapter.cellzero \
           or self.to_cell(val) == self.adapter.addablecellzero or change.old == traitlets.Undefined:
            return
        if not hasattr(self.adapter, 'remove_cell'):
            raise TypeError("Cannot remove cell from this object.")
        if not self.adapter.remove_cell or self.adapter.remove_cell.__func__.__class__ is AbstractMethod:
            return # Method not implemented or deliberately set to None
        if hasattr(self.adapter.remove_cell, '_optional') and self.adapter.remove_cell._optional: # Not implemented
            raise Exception("Removing cells is not implemented for this object.")
        obj = copy(self.value) # For your pet objects, don't forget to implement __copy__
        result = self.adapter.remove_cell(obj, pos, dirty=self.dirty)
        if issubclass(result.__class__, BaseException): # Removing cell was impossible
            if pos in self.addable_cells() or (pos in self.cells and val == self.cells[pos]) \
               and self.dirty.keys() == [pos]: # Rollback
                self.reset_dirty()
            else: # Keep temporary substraction for later
                self.set_dirty(pos, val, result)
            return
        self.set_value(result)

    def append_row(self, r=None):
        r"""
        Append a row to editor value.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: e = GridViewEditor(m)
            sage: e.donottrack = False
            sage: e.value
            [ 1  7  1]
            [ 0  0  3]
            [ 0 -1  2]
            [ 1  0 -3]
            sage: e.append_row((1,2,3))
            sage: e.value
            [ 1  7  1]
            [ 0  0  3]
            [ 0 -1  2]
            [ 1  0 -3]
            [ 1  2  3]
            sage: from sage.combinat.tableau import Tableau
            sage: e = GridViewEditor(Tableau([[1, 4, 7, 8, 9, 10, 11], [2, 5, 13], [3, 6], [12, 15], [14]]))
            sage: e.donottrack = False
            sage: e.value
            [[1, 4, 7, 8, 9, 10, 11], [2, 5, 13], [3, 6], [12, 15], [14]]
            sage: e.append_row((1,2,3))
            Traceback (most recent call last):
            ...
            TypeError: 'NotImplementedType' object is not callable
        """
        if self.donottrack:
            return
        if not hasattr(self.adapter, 'append_row'):
            raise TypeError("Cannot append row to this object.")
        obj = copy(self.value)
        obj = self.adapter.append_row(obj, r)
        self.value = obj # Will call the observer

    def insert_row(self, index, r=None):
        r"""
        Insert a row into editor value.
        """
        if self.donottrack:
            return
        if not hasattr(self.adapter, 'insert_row'):
            raise TypeError("Cannot insert row to this object.")
        obj = copy(self.value)
        obj = self.adapter.insert_row(obj, index, r)
        self.value = obj # Will call the observer

    def remove_row(self, index=None):
        r"""
        Remove a row from editor value.
        """
        if self.donottrack:
            return
        if not hasattr(self.adapter, 'remove_row'):
            raise TypeError("Cannot remove row from this object.")
        obj = copy(self.value)
        obj = self.adapter.remove_row(obj, index)
        self.value = obj # Will call the observer

    def append_column(self, c=None):
        r"""
        Append a column to editor value.

        TESTS ::

            sage: from sage_combinat_widgets import GridViewEditor
            sage: from sage.matrix.matrix_space import MatrixSpace
            sage: S = MatrixSpace(ZZ, 4,3)
            sage: m = S.matrix([1,7,1,0,0,3,0,-1,2,1,0,-3])
            sage: e = GridViewEditor(m)
            sage: e.donottrack = False
            sage: e.value
            [ 1  7  1]
            [ 0  0  3]
            [ 0 -1  2]
            [ 1  0 -3]
            sage: e.append_column((1,2,3))
            sage: e.value
            [ 1  7  1  1]
            [ 0  0  3  2]
            [ 0 -1  2  3]
            [ 1  0 -3  0]
            sage: from sage.combinat.tableau import Tableau
            sage: e = GridViewEditor(Tableau([[1, 4, 7, 8, 9, 10, 11], [2, 5, 13], [3, 6], [12, 15], [14]]))
            sage: e.donottrack = False
            sage: e.value
            [[1, 4, 7, 8, 9, 10, 11], [2, 5, 13], [3, 6], [12, 15], [14]]
            sage: e.append_column((1,2,3))
            Traceback (most recent call last):
            ...
            TypeError: 'NotImplementedType' object is not callable
        """
        if self.donottrack:
            return
        if not hasattr(self.adapter, 'append_column'):
            raise TypeError("Cannot append column to this object.")
        obj = copy(self.value)
        obj = self.adapter.append_column(obj, c)
        self.value = obj # Will call the observer

    def insert_column(self, index, c=None):
        r"""
        Insert a column into editor value.
        """
        if self.donottrack:
            return
        if not hasattr(self.adapter, 'insert_column'):
            raise TypeError("Cannot insert column to this object.")
        obj = copy(self.value)
        obj = self.adapter.insert_column(obj, index, c)
        self.value = obj # Will call the observer

    def remove_column(self, index=None):
        r"""
        Remove a column from editor value.
        """
        if self.donottrack:
            return
        if not hasattr(self.adapter, 'remove_column'):
            raise TypeError("Cannot remove column from this object.")
        obj = copy(self.value)
        obj = self.adapter.remove_column(obj, index)
        self.value = obj # Will call the observer



# -*- coding: utf-8 -*-
r"""
Grid View Adapter for partitions

**Grid View partition operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~PartitionGridViewAdapter.cell_to_display` | Static method for typecasting cell content to widget display value
    :meth:`~PartitionGridViewAdapter.display_to_cell` | Instance method for typecasting widget display value to cell content
    :meth:`~PartitionGridViewAdapter.compute_cells` | Compute partition cells as a dictionary { coordinate pair : Integer }
    :meth:`~PartitionGridViewAdapter.from_cells` | Create a new partition from a cells dictionary
    :meth:`~PartitionGridViewAdapter.get_cell` | Get the partition cell content
    :meth:`~PartitionGridViewAdapter.addable_cells` | List addable cells
    :meth:`~PartitionGridViewAdapter.removable_cells` | List removable cells
    :meth:`~PartitionGridViewAdapter.add_cell` | Add a cell
    :meth:`~PartitionGridViewAdapter.remove_cell` | Remove a cell

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
from sage.combinat.partition import Partition
from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
from six import text_type

class PartitionGridViewAdapter(GridViewAdapter):
    r"""
    Grid view adapter for partitions.

    ATTRIBUTES::
        * ``objclass`` -- Partition
        * ``celltype`` -- bool
        * ``cellzero`` -- False
    """
    objclass = Partition
    constructorname = 'Partition'
    celltype = bool
    cellzero = False

    @staticmethod
    def cell_to_display(cell_content, display_type=bool):
        r"""
        From object cell content
        to widget display value.

        TESTS ::

            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: PartitionGridViewAdapter.cell_to_display(True)
            True
            sage: from six import text_type
            sage: PartitionGridViewAdapter.cell_to_display("my string", text_type)
            ''
        """
        if display_type == text_type:
            return ''
        return cell_content

    def display_to_cell(self, display_value, display_type=bool):
        r"""
        From widget cell value
        to object display content

        TESTS ::

            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: pa = PartitionGridViewAdapter()
            sage: pa.display_to_cell(True)
            True
            sage: pa.display_to_cell('')
            False
        """
        if not display_value or display_type == text_type:
            return self.cellzero
        return display_value

    @staticmethod
    def compute_cells(obj):
        r"""
        From a partition,
        return a dictionary { coordinates pair : Integer }

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: p = Partition([3, 2, 1, 1])
            sage: PartitionGridViewAdapter.compute_cells(p)
            {(0, 0): False,
            (0, 1): False,
            (0, 2): False,
            (1, 0): False,
            (1, 1): False,
            (2, 0): False,
            (3, 0): False}
        """
        return {(i,j):False for (i,j) in obj.cells()}

    @classmethod
    def from_cells(cls, cells={}):
        r"""
        From a dictionary { coordinates pair : Integer }
        return a corresponding partition

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: PartitionGridViewAdapter.from_cells({(0, 0): False, (0, 1): False, (0, 2): True, (0, 3): False, (1, 0): False, (2, 0): True})
            [4, 1, 1]
        """
        partition_elements = [
            len([(i, pos[1]) for pos in cells if pos[0] == i]) for i in range(max(pos[0] for pos in cells) + 1)]
        try:
            return cls.objclass(partition_elements)
        except:
            raise TypeError(
                "This object is not compatible with this adapter (%s, for %s objects)" % (cls, cls.objclass))

    @staticmethod
    def get_cell(obj, pos):
        r"""
        Get cell value

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: p = Partition([6, 5, 2, 1])
            sage: PartitionGridViewAdapter.get_cell(p, (1, 1))
            False
            sage: PartitionGridViewAdapter.get_cell(p, (1, 6))
            Traceback (most recent call last):
            ...
            ValueError: Cell '(1, 6)' not in partition.
        """
        try:
            assert pos[0] < len(obj) and pos[1] < obj[pos[0]]
        except:
            raise ValueError("Cell '%s' not in partition." % str(pos))
        return False

    def set_cell(self, obj, pos, val, dirty={}, constructorname=''):
        r"""
        From a partition `obj`, a position (pair of coordinates) `pos` and a value `val`,
        return a new partition with a modified cell at position `pos`.
        Actually remove the cell if it's removable, otherwise return the same partition.

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: p = Partition([6, 5, 2, 1])
            sage: pa = PartitionGridViewAdapter()
            sage: pa.set_cell(p, (1,2), True)
            [6, 5, 2, 1]
            sage: pa.set_cell(p, (1,4), True)
            [6, 4, 2, 1]
        """
        if pos in self.removable_cells(obj):
            return self.remove_cell(obj, pos, dirty)
        return obj

    @staticmethod
    def addable_cells(obj):
        r"""
        List object addable cells

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: p = Partition([6, 5, 2, 1])
            sage: PartitionGridViewAdapter.addable_cells(p)
            [(0, 6), (1, 5), (2, 2), (3, 1), (4, 0)]
        """
        return obj.outside_corners()

    @staticmethod
    def removable_cells(obj):
        r"""
        List object removable cells

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: p = Partition([6, 5, 2, 1])
            sage: PartitionGridViewAdapter.removable_cells(p)
            [(0, 5), (1, 4), (2, 1), (3, 0)]
        """
        return obj.corners()

    def add_cell(self, obj, pos, val=None, dirty={}):
        r"""
        Add cell

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: p = Partition([6, 5, 2, 1])
            sage: pa = PartitionGridViewAdapter()
            sage: pa.add_cell(p, (2, 2))
            [6, 5, 3, 1]
            sage: pa.add_cell(p, (4, 0), 42)
            [6, 5, 2, 1, 1]
            sage: pa.add_cell(p, (2, 0))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(2, 0)' is not addable.
        """
        if not pos in self.addable_cells(obj):
            raise ValueError("Cell position '%s' is not addable." % str(pos))
        try:
            return obj.add_cell(pos[0])
        except Exception as e:
            return e

    def remove_cell(self, obj, pos, dirty={}):
        r"""
        Remove cell

        TESTS ::

            sage: from sage.combinat.partition import Partition
            sage: from sage_widget_adapters.combinat.partition_grid_view_adapter import PartitionGridViewAdapter
            sage: p = Partition([6, 5, 2, 1])
            sage: pa = PartitionGridViewAdapter()
            sage: pa.remove_cell(p, (2, 1))
            [6, 5, 1, 1]
            sage: pa.remove_cell(p, (1, 1))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(1, 1)' is not removable.
        """
        if not pos in self.removable_cells(obj):
            raise ValueError("Cell position '%s' is not removable." % str(pos))
        try:
            return obj.remove_cell(pos[0])
        except Exception as e:
            return e



# -*- coding: utf-8 -*-
r"""
Grid View Adapter for skew tableaux

**Grid View skew tableau operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~SkewTableauGridViewAdapter.compute_cells` | Compute skew tableau cells as a dictionary { coordinate pair : Integer }
    :meth:`~SkewTableauGridViewAdapter.from_cells` | Create a new skew tableau from a cells dictionary
    :meth:`~SkewTableauGridViewAdapter.addable_cells` | List addable cells
    :meth:`~SkewTableauGridViewAdapter.removable_cells` | List removable cells
    :meth:`~SkewTableauGridViewAdapter.add_cell` | Add a cell
    :meth:`~SkewTableauGridViewAdapter.remove_cell` | Remove a cell

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
from sage.combinat.skew_tableau import SkewTableau
from sage.rings.integer import Integer
from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter


class SkewTableauGridViewAdapter(GridViewAdapter):
    r"""
    Grid view adapter for skew tableaux.

    ATTRIBUTES::
        * ``objclass`` -- SkewTableau
        * ``celltype`` -- Integer
        * ``cellzero`` -- Integer(0)
    """
    objclass = SkewTableau
    constructorname = 'SkewTableau'
    celltype = Integer # i.e. sage.rings.integer.Integer
    cellzero = Integer(0)

    @staticmethod
    def compute_cells(obj):
        r"""
        From a skew tableau,
        return a dictionary { coordinates pair : Integer }
        TESTS ::

            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: from sage_widget_adapters.combinat.skew_tableau_grid_view_adapter import SkewTableauGridViewAdapter
            sage: st = SkewTableau([[None, None, 1, 2], [None, 1], [4]])
            sage: SkewTableauGridViewAdapter.compute_cells(st)
            {(0, 2): 1, (0, 3): 2, (1, 1): 1, (2, 0): 4}
        """
        return {(i,j):obj[i][j] for (i,j) in obj.cells()}

    @classmethod
    def from_cells(cls, cells={}):
        r"""
        From a dictionary { coordinates pair : Integer }
        return a corresponding skew tableau
        TESTS ::

            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: from sage_widget_adapters.combinat.skew_tableau_grid_view_adapter import SkewTableauGridViewAdapter
            sage: SkewTableauGridViewAdapter.from_cells({(0, 1): 2, (0, 3): 6, (1, 0): 3, (2, 0): 4})
            [[None, 2, None, 6], [3], [4]]
        """
        rows = []
        for i in range(max(pos[0] for pos in cells) + 1):
            rows.append([None] * (max(pos[1] for pos in cells if pos[0] == i) + 1))
        for pos in cells:
            rows[pos[0]][pos[1]] = cells[pos]
        try:
            return cls.objclass(rows)
        except:
            raise TypeError(
                "This object is not compatible with this adapter (%s, for %s objects)" % (cls, cls.objclass))

    @staticmethod
    def addable_cells(obj):
        r"""
        List object addable cells
        TESTS ::

            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: from sage_widget_adapters.combinat.skew_tableau_grid_view_adapter import SkewTableauGridViewAdapter
            sage: st = SkewTableau([[None, None, None, 1], [None, None, 2], [None, 1], [4]])
            sage: SkewTableauGridViewAdapter.addable_cells(st)
            [(0, 2), (1, 1), (2, 0), (0, 4), (1, 3), (2, 2), (3, 1), (4, 0)]
        """
        return obj.shape().inner().corners() + obj.shape().outer().outside_corners()

    @staticmethod
    def removable_cells(obj):
        r"""
        List object removable cells
        TESTS ::

            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: from sage_widget_adapters.combinat.skew_tableau_grid_view_adapter import SkewTableauGridViewAdapter
            sage: st = SkewTableau([[None, None, None, 1, 2, 6], [None, None, 3, 4], [None, 1], [5]])
            sage: SkewTableauGridViewAdapter.removable_cells(st)
            [(0, 5), (1, 3), (2, 1), (3, 0), (0, 3), (1, 2)]
        """
        ret = obj.shape().outer().corners()
        for c in obj.shape().inner().outside_corners():
            if not c in ret:
                ret.append(c)
        return ret

    def add_cell(self, obj, pos, val, dirty={}):
        r"""
        Add cell.

        TESTS ::

            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: from sage_widget_adapters.combinat.skew_tableau_grid_view_adapter import SkewTableauGridViewAdapter
            sage: st = SkewTableau([[None, None, None, 2], [None, 1, 1], [1], [4]])
            sage: sta = SkewTableauGridViewAdapter()
            sage: sta.add_cell(st, (0, 2), 1)
            [[None, None, 1, 2], [None, 1, 1], [1], [4]]
            sage: sta.add_cell(st, (4, 0), 7)
            [[None, None, None, 2], [None, 1, 1], [1], [4], [7]]
            sage: sta.add_cell(st, (1, 1), 9)
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(1, 1)' is not addable.
        """
        if not pos in self.addable_cells(obj):
            raise ValueError("Cell position '%s' is not addable." % str(pos))
        sl = obj.to_list()
        sl = self.make_dirty(sl, dirty)
        if pos[0] >= len(obj):
            sl.append([val])
        elif pos in obj.shape().outer().outside_corners():
            sl[pos[0]].append(val)
        else:
            sl[pos[0]][pos[1]] = val
        return self._validate(sl)

    def remove_cell(self, obj, pos, dirty={}):
        r"""
        Remove cell.

        TESTS ::

            sage: from sage.combinat.skew_tableau import SkewTableau
            sage: from sage_widget_adapters.combinat.skew_tableau_grid_view_adapter import SkewTableauGridViewAdapter
            sage: st = SkewTableau([[None, None, None, 1, 2, 6], [None, None, 3, 4], [None, 1], [5]])
            sage: sta = SkewTableauGridViewAdapter()
            sage: sta.remove_cell(st, (0, 0))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(0, 0)' is not removable.
            sage: sta.remove_cell(st, (0, 3))
            [[None, None, None, None, 2, 6], [None, None, 3, 4], [None, 1], [5]]
            sage: sta.remove_cell(st, (2, 1))
            [[None, None, None, 1, 2, 6], [None, None, 3, 4], [None], [5]]
            sage: st = SkewTableau([[None, None, 1, 2, 3], [None, 1], [4]])
            sage: sta.remove_cell(st, (0, 4))
            [[None, None, 1, 2], [None, 1], [4]]
        """
        if not pos in self.removable_cells(obj):
            raise ValueError("Cell position '%s' is not removable." % str(pos))
        sl = obj.to_list()
        sl = self.make_dirty(sl, dirty)
        if len(sl[pos[0]]) == 1:
            del(sl[pos[0]])
        elif pos in obj.shape().outer().corners():
            sl[pos[0]].pop()
        else:
            sl[pos[0]][pos[1]] = None
        return self._validate(sl)



#!/usr/bin/env python
# coding: utf-8

from flipping_aztecdiamond import DominoGeometry, DominosAdapter
from sage_combinat_widgets.grid_view_widget import GridViewWidget, ButtonCell, BlankButton, styled_button_cell
from ipywidgets import Layout
from traitlets import dlink, HasTraits, Bool
from contextlib import contextmanager

smallblyt = Layout(width='12px',height='12px', margin='0', padding='0')


class ddlink(dlink):
    """Double directional link with logic = or/and/none.
    Double_source is a tuple of (source, target) tuples
    Usage: b1 = Button(description = 'B1')
           b2 = Button(description = 'B2')
           b3 = Button(description = 'B3')
           dl = ddlink(((b1, 'value'), (b2, 'value')), (b3, 'value'))
    """
    updating = False

    def __init__(self, double_source, target, logic=None, set_at_init=True):
        self.validate_tuple(target)
        self.logic = logic
        for source in double_source:
            self.validate_tuple(source)
        try:
            if set_at_init:
                if self.logic == 'and':
                    setattr(target[0], target[1],
                            getattr(double_source[0][0], double_source[0][1]) \
                            and getattr(double_source[1][0], double_source[1][1]))
                else:
                    setattr(target[0], target[1],
                            getattr(double_source[0][0], double_source[0][1]) \
                            or getattr(double_source[1][0], double_source[1][1]))
            else:
                pass
        finally:
            for source in double_source:
                source[0].observe(self._update, names=source[1])
                source[0].observe(self._update, names=source[1])
        self.target = target
        self.double_source = double_source
        self.intermediate_value = False # False / True / source.tooltip

    def __repr__(self):
        if self.double_source or self.target:
            return "A double directional link from sources=%s to target='%s'" % (self.double_source, self.target)
        return "None"

    def validate_tuple(self, t):
        if not len(t) == 2:
            raise TypeError("Each linked traitlet must be specified as (HasTraits, 'trait_name'), not %r" % t)
        obj, trait_name = t
        if not isinstance(obj, HasTraits):
            raise TypeError("Each object must be HasTraits, not %r" % type(obj))
        if not trait_name in obj.traits():
            raise TypeError("%r has no trait %r" % (obj, trait_name))

    @contextmanager
    def _busy_updating(self):
        self.updating = True
        try:
            yield
        finally:
            self.updating = False

    def _update(self, change):
        if self.updating:
            return
        with self._busy_updating():
            if self.logic == 'and':
                if self.intermediate_value == False: # aucun bouton pressé avant
                    self.intermediate_value = change.new
                elif self.intermediate_value == True: # 2 boutons pressés avant
                    setattr(self.target[0], self.target[1], change.new)
                    if change.new == False:
                        # prendre le tooltip de l'autre bouton
                        for s in self.double_source:
                            if s[0].tooltip != change.owner.tooltip:
                                self.intermediate_value = s[0].tooltip
                elif self.intermediate_value != change.owner.tooltip: # l'autre bouton a été pressé avant
                    setattr(self.target[0], self.target[1], change.new)
                    self.intermediate_value = change.new
            elif self.logic == 'or':
                if change.new == True:
                    setattr(self.target[0], self.target[1], True)
                    if self.intermediate_value == False:
                        self.intermediate_value = change.owner.tooltip
                    elif self.intermediate_value == True:
                        pass
                    elif self.intermediate_value != change.owner.tooltip:
                        self.intermediate_value = True
                else:
                    if self.intermediate_value != True:
                        setattr(self.target[0], self.target[1], False)
                    self.intermediate_value = False # FIXME

            else:
                setattr(self.target[0], self.target[1], change.new)

    def unlink(self):
        for source in self.double_source:
            source[0].unobserve(self._update, names=source[1])
        self.double_source, self.target = None, None

class Domino(HasTraits):
    r"""Objet non représenté en lui-même, les 2
    boutons qu'il contient étant, eux, des widgets"""
    value = Bool()

    def __init__(self, parent, b1, b2):
        """A domino has a parent widget and is made of 2 buttons"""
        super(Domino, self).__init__()
        self.geometry = DominoGeometry(b1.position, b2.position)
        self.parent = parent
        self.key = None
        self.first = b1
        self.second = b2
        self.buttons = (b1,b2)
        b1.link = dlink((b1, 'value'), (b2, 'value'))
        b2.link = dlink((b2, 'value'), (b1, 'value'))
        self.link = None
        self.direction = None
        self.orientation = None
        self.compute()

    def __repr__(self):
        if self.value:
            return repr(self.first) + ' -> ' + repr(self.second) + " PRESSED"
        return repr(self.first) + ' -> ' + repr(self.second)

    def compute(self):
        """Compute buttons relative positions.
        Create double directional link from both buttons"""
        self.geometry.compute()
        self.link = ddlink(((self.first, 'value'), (self.second, 'value')), (self, 'value'), logic='and', set_at_init=False) # Fresh ddlink
        if self.geometry.orientation == 1:
            self.key = self.first.position
        else:
            self.key = self.second.position
        for cl in ['left','right','top','bottom']:
            self.first.remove_class(cl)
            self.second.remove_class(cl)
        if self.geometry.direction == 'horizontal':
            if self.geometry.orientation == 1:
                self.first.add_class('left')
                self.second.add_class('right')
            elif self.geometry.orientation == -1:
                self.first.add_class('right')
                self.second.add_class('left')
        elif self.geometry.direction == 'vertical':
            if self.geometry.orientation == 1:
                self.first.add_class('top')
                self.second.add_class('bottom')
            elif self.geometry.orientation == -1:
                self.first.add_class('bottom')
                self.second.add_class('top')

    def is_pressed(self):
        """Is the domino pressed?"""
        return self.value

    def reset(self):
        """Full domino reset"""
        self.value = False
        self.link.unlink()

    def flip(self, other):
        """Flip self with some neighboring domino"""
        self.reset()
        other.reset()
        self.geometry.flip(other.geometry)
        self.compute()
        other.compute()

def make_cell_widget_class_index(g):
    def cell_widget_class_index(pos):
        def calc_index_for_domino(d):
            if d.direction == 'horizontal':
                if not d.parity:
                    return 1
                else:
                    return 2
            else:
                if not d.parity:
                    return 3
                else:
                    return 4
        d = g.domino_for_position(pos)
        if d:
            return calc_index_for_domino(d)
        return 0
    return cell_widget_class_index

class DominosWidget(GridViewWidget):
    """A widget with dominos"""

    def __init__(self, g):
        r"""
        Init a flipping dominos widget
        with flipping aztec diamond graph `g`
        """
        super(DominosWidget, self).__init__(g, adapter = DominosAdapter(),
                                            cell_layout = smallblyt,
                                            cell_widget_classes=[styled_button_cell(),
                                                                 styled_button_cell(style_name='b1'),
                                                                 styled_button_cell(style_name='b2'),
                                                                 styled_button_cell(style_name='b3'),
                                                                 styled_button_cell(style_name='b4'),
                                            ],
                                            cell_widget_class_index=make_cell_widget_class_index(g),
                                            blank_widget_class = BlankButton)

    def draw(self):
        self.dominos = {}
        super(DominosWidget, self).draw()
        self.apply_matching(self.value.matching)

    def match(self, b1, b2):
        """Match buttons b1 and b2, that is: create a domino"""
        try:
            assert issubclass(b1.__class__, ButtonCell) and issubclass(b2.__class__, ButtonCell)
        except:
            raise Exception('This method requires 2 buttons!! b1 = %s, b2 = %s' % (b1, b2) )
        """Create a domino and let it do the work. NB: the key should always be top-left."""
        d = Domino(self, b1, b2)
        self.dominos[d.key] = d

    def reset(self):
        """Clear dominos and reset every button"""
        self.dominos = {}

    def apply_matching(self, matching):
        """Apply a matching"""
        for d in matching:
            self.match(self.children[d.first[0]].children[d.first[1]],
                       self.children[d.second[0]].children[d.second[1]])



from __future__ import print_function, absolute_import
# Add the import for which you want to give a direct access
from .grid_view_editor import GridViewEditor
from .grid_view_widget import BlankCell, DisabledButtonCell, BlankButton, styled_button_cell, styled_push_button, GridViewWidget
from sage_widget_adapters import *



#!/usr/bin/env python
# coding: utf-8

from sage.graphs.graph import Graph
from sage_widget_adapters.graphs.graph_grid_view_adapter import GraphGridViewAdapter

class DominoGeometry:
    r"""
    The geometry of the domino."""
    def __init__(self, first, second):
        """A domino is a pair of pairs"""
        self.first = first
        self.second = second
        self.direction = None
        self.orientation = None
        self.compute()

    def __repr__(self):
        return "DominoGeometry %s -> %s" % (self.first, self.second)

    def __eq__(self, other):
        assert issubclass(other.__class__, DominoGeometry)
        return self.first == other.first and self.second == other.second

    def __lt__(self, other):
        assert issubclass(other.__class__, DominoGeometry)
        if other.direction != self.direction:
            return False
        return (self.first < other.first and self.second < other.second) \
            or (self.first < other.second and self.second < other.first)

    def compute(self):
        r"""Compute domino direction and orientation."""
        if self.first[0] == self.second[0]: # same row
            self.direction = 'horizontal'
            if self.first[1] + 1 == self.second[1]:
                self.orientation = 1 # left to right
            elif self.first[1] == self.second[1] + 1:
                self.orientation = -1 # right to left
        elif self.first[1] == self.second[1]: # same column
            self.direction = 'vertical'
            if self.first[0] + 1 == self.second[0]:
                self.orientation = 1 # top to bottom
            elif self.first[0] == self.second[0] + 1:
                self.orientation = -1 # bottom to top
        if self.orientation == 1:
            self.parity = (self.first[0]%2 + self.first[1]%2)%2
        elif self.orientation == -1:
            self.parity = (self.second[0]%2 + self.second[1]%2)%2

    def reverse(self):
        return DominoGeometry(self.second, self.first)

    def neighbors(self):
        r"""
        Return list of parallel neighbouring matches
        Note: we consider only horizontal or vertical matches"""
        if self.direction == 'horizontal':
            return [DominoGeometry((self.first[0] + 1, self.first[1]), (self.second[0] + 1, self.second[1])),
                    DominoGeometry((self.first[0] - 1, self.first[1]), (self.second[0] - 1, self.second[1]))]
        else:
            return [DominoGeometry((self.first[0], self.first[1] + 1), (self.second[0], self.second[1] + 1)),
                    DominoGeometry((self.first[0], self.first[1] - 1), (self.second[0], self.second[1] - 1))]

    def flip(self, other):
        """Flip self with some neighboring domino"""
        if self.orientation * other.orientation == 1: # Same orientation
            self.second, other.first = other.first, self.second
        else:
            self.second, other.second = other.second, self.second
        self.compute()
        other.compute()

class FlippingAztecDiamond(Graph):
    def __init__(self, n, matching=()):
        try:
            assert n > 0 and n.is_integer()
        except:
            raise TypeError("We need a positive integer for the graph order.")
        N = 2 * n
        positions = ((i, j) for i in range(N) for j in range(N) if i - n <= j <= n + i and n - 1 - i <= j <= 3 * n - i - 1)
        super(FlippingAztecDiamond, self).__init__(name="Flipping Aztec Diamond Graph of order {}".format(n))
        self.aztec_order = n
        self.add_vertices(positions)
        self.add_edges((pos,(pos[0],pos[1]+1)) for pos in positions if (pos[0],pos[1]+1) in positions)
        self.add_edges((pos,(pos[0]+1,pos[1])) for pos in positions if (pos[0]+1,pos[1]) in positions)
        try:
            for t in matching:
                assert (t[0][0] == t[1][0] and (t[0][1] + 1 == t[1][1] or t[0][1] == t[1][1] + 1) or \
                        t[0][1] == t[1][1] and (t[0][0] + 1 == t[1][0] or t[0][0] == t[1][0] + 1))
        except:
            raise TypeError("This matching is not suitable for a flipping aztec diamond (tuple {} is not valid)." . format(t))
        self.matching = self.apply_matching(matching) # dominos with only horizontal or vertical consecutive matches

    def __copy__(self):
        return FlippingAztecDiamond.__init__(self, self.aztec_order, ((m.first, m.second) for m in self.matching))

    def apply_matching(self, matching):
        self.matching = [DominoGeometry(t[0], t[1]) for t in matching]

    def domino_for_position(self, pos):
        try:
            assert type(pos) is type((0,))
            assert pos in self.vertices()
        except:
            raise TypeError("Argument `pos` must be a tuple and a graph vertex")
        for d in self.matching:
            if d.first == pos or d.second == pos:
                return d

    @staticmethod
    def flip(d1, d2):
        """d1 and d2 are dominos"""
        if d1 < d2:
            d1.flip(d2)
        else:
            d2.flip(d1)

class DominosAdapter(GraphGridViewAdapter):

    remove_cell = None

    def set_cell(self, obj, pos, val=True, dirty={}):
        r"""
        When we click on a graph cell,
        we prepare a possible flip
        or we try to complete the flip if it has been prepared previously
        """
        # Find out the relevant matching for 'pos'
        d1 = obj.domino_for_position(pos)
        if dirty: # if i'm a neighbor, then flip and return a new obj ; else return an error
            # Consider the relevant matching(s)
            for p in dirty:
                if not dirty[p]: # check this position is pressed
                    continue
                d2 = obj.domino_for_position(p)
                if d2 == d1:
                    continue
                if d2 in d1.neighbors():
                    # Do the flip
                    obj.flip(d1, d2)
                    return obj
            return Exception("Please select a second domino!")
        else:
            return Exception("Please select a second domino!")



# -*- coding: utf-8 -*-
r"""
Grid View Adapter for skew partitions

**Grid View skew partition operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~SkewPartitionGridViewAdapter.cell_to_display` | Static method for typecasting cell content to widget display value
    :meth:`~SkewPartitionGridViewAdapter.display_to_cell` | Instance method for typecasting widget display value to cell content
    :meth:`~SkewPartitionGridViewAdapter.compute_cells` | Compute skew partition cells as a dictionary { coordinate pair : Integer }
    :meth:`~SkewPartitionGridViewAdapter.from_cells` | Create a new skew partition from a cells dictionary
    :meth:`~SkewPartitionGridViewAdapter.get_cell` | Get the skew partition cell content
    :meth:`~SkewPartitionGridViewAdapter.addable_cells` | List addable cells
    :meth:`~SkewPartitionGridViewAdapter.removable_cells` | List removable cells
    :meth:`~SkewPartitionGridViewAdapter.add_cell` | Add a cell
    :meth:`~SkewPartitionGridViewAdapter.remove_cell` | Remove a cell

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
from sage.combinat.skew_partition import SkewPartition
from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter
from six import text_type


class SkewPartitionGridViewAdapter(GridViewAdapter):
    r"""
    Grid view adapter for skew partitions.

    ATTRIBUTES::
        * ``objclass`` -- SkewPartition
        * ``celltype`` -- bool
        * ``cellzero`` -- False
    """
    objclass = SkewPartition
    celltype = bool
    cellzero = False

    @staticmethod
    def cell_to_display(cell_content, display_type=bool):
        r"""
        From object cell content
        to widget display value.

        TESTS ::

            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: SkewPartitionGridViewAdapter.cell_to_display(True)
            True
            sage: from six import text_type
            sage: SkewPartitionGridViewAdapter.cell_to_display("my string", text_type)
            ''
        """
        if display_type == text_type:
            return ''
        return cell_content

    def display_to_cell(self, display_value, display_type=bool):
        r"""
        From widget cell value
        to object display content

        TESTS ::

            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: pa = SkewPartitionGridViewAdapter()
            sage: pa.display_to_cell(True)
            True
            sage: pa.display_to_cell('')
            False
        """
        if not display_value or display_type == text_type:
            return self.cellzero
        return display_value

    @staticmethod
    def compute_cells(obj):
        r"""
        From a skew partition,
        return a dictionary { coordinates pair : Integer }

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: sp = SkewPartition([[4, 2, 1],[2, 1]])
            sage: SkewPartitionGridViewAdapter.compute_cells(sp)
            {(0, 2): False, (0, 3): False, (1, 1): False, (2, 0): False}
        """
        return {(i,j):False for (i,j) in obj.cells()}

    @classmethod
    def from_cells(cls, cells={}):
        r"""
        From a dictionary { coordinates pair : Integer }
        return a corresponding skew partition.

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: SkewPartitionGridViewAdapter.from_cells({(0, 2): False, (0, 3): False, (1, 1): False, (2, 0): False})
            [4, 2, 1] / [2, 1]
        """
        height = max([pos[0] for pos in cells]) + 1
        outer = [max([pos[1] for pos in cells if pos[0]==i]) + 1 for i in range(height)]
        inner = [min([pos[1] for pos in cells if pos[0]==i]) for i in \
                 range(height) if min([pos[1] for pos in cells if pos[0]==i]) > 0]
        try:
            return cls.objclass([outer,inner])
        except:
            raise TypeError(
                "This object is not compatible with this adapter (%s, for %s objects)" % (cls, cls.objclass))

    @staticmethod
    def get_cell(obj, pos):
        r"""
        Get cell value

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: sp = SkewPartition([[4, 2, 1],[2, 1]])
            sage: SkewPartitionGridViewAdapter.get_cell(sp, (1, 1))
            False
            sage: SkewPartitionGridViewAdapter.get_cell(sp, (1, 0))
            Traceback (most recent call last):
            ...
            ValueError: Cell '(1, 0)' not in object.
        """
        try:
            assert pos[0] < len(obj) and pos[1] < obj.outer()[pos[0]] and pos[1] >= obj.inner()[pos[0]]
        except:
            raise ValueError("Cell '%s' not in object." % str(pos))
        return False

    def set_cell(self, obj, pos, val, dirty={}, constructorname=''):
        r"""
        From a partition `obj`, a position (pair of coordinates) `pos` and a value `val`,
        return a new partition with a modified cell at position `pos`.
        Actually remove the cell if it's removable, otherwise return the same partition.

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: sp = SkewPartition([[7, 4, 2, 1],[2, 1, 1]])
            sage: spa = SkewPartitionGridViewAdapter()
            sage: spa.set_cell(sp, (1,6), True)
            [7, 4, 2, 1] / [2, 1, 1]
            sage: spa.set_cell(sp, (1,3), True)
            [7, 3, 2, 1] / [2, 1, 1]
        """
        if pos in self.removable_cells(obj):
            return self.remove_cell(obj, pos, dirty)
        return obj

    @staticmethod
    def addable_cells(obj):
        r"""
        List object addable cells

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: sp = SkewPartition([[4, 2, 1],[2, 1]])
            sage: SkewPartitionGridViewAdapter.addable_cells(sp)
            [(0, 1), (1, 0), (0, 4), (1, 2), (2, 1), (3, 0)]
        """
        return obj.inner().corners() + obj.outer().outside_corners()

    @staticmethod
    def removable_cells(obj):
        r"""
        List object removable cells

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: SkewPartitionGridViewAdapter.removable_cells(SkewPartition([[4, 2, 1],[2, 1]]))
            [(0, 2), (1, 1), (2, 0), (0, 3)]
            sage: SkewPartitionGridViewAdapter.removable_cells(SkewPartition([[7, 4, 2, 1],[2, 1, 1]]))
            [(0, 2), (1, 1), (3, 0), (0, 6), (1, 3), (2, 1)]
        """
        ret = obj.inner().outside_corners()
        for c in obj.outer().corners():
            if not c in ret:
                ret.append(c)
        return ret

    def add_cell(self, obj, pos, val=None, dirty={}):
        r"""
        Add cell

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: sp = SkewPartition([[7, 4, 2, 1],[2, 1, 1]])
            sage: spa = SkewPartitionGridViewAdapter()
            sage: spa.add_cell(sp, (0, 7))
            [8, 4, 2, 1] / [2, 1, 1]
            sage: spa.add_cell(sp, (2, 0))
            [7, 4, 2, 1] / [2, 1]
            sage: spa.add_cell(sp, (4, 0))
            [7, 4, 2, 1, 1] / [2, 1, 1]
            sage: spa.add_cell(sp, (2, 3))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(2, 3)' is not addable.
        """
        if not pos in self.addable_cells(obj):
            raise ValueError("Cell position '%s' is not addable." % str(pos))
        try:
            if pos in obj.outer().outside_corners():
                return self.objclass([obj.outer().add_cell(pos[0]), obj.inner()])
            else:
                return self.objclass([obj.outer(), obj.inner().remove_cell(pos[0])])
        except:
            raise ValueError("Error adding cell %s to %s" % (pos, self.objclass))

    def remove_cell(self, obj, pos, dirty={}):
        r"""
        Remove cell

        TESTS ::

            sage: from sage.combinat.skew_partition import SkewPartition
            sage: from sage_widget_adapters.combinat.skew_partition_grid_view_adapter import SkewPartitionGridViewAdapter
            sage: sp = SkewPartition([[7, 4, 2, 1],[2, 1, 1]])
            sage: spa = SkewPartitionGridViewAdapter()
            sage: spa.remove_cell(sp, (0, 6))
            [6, 4, 2, 1] / [2, 1, 1]
            sage: spa.remove_cell(sp, (1, 1))
            [7, 4, 2, 1] / [2, 2, 1]
            sage: spa.remove_cell(sp, (1, 2))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(1, 2)' is not removable.
        """
        if not pos in self.removable_cells(obj):
            raise ValueError("Cell position '%s' is not removable." % str(pos))
        try:
            if pos in obj.outer().corners():
                return self.objclass([obj.outer().remove_cell(pos[0]), obj.inner()])
            else:
                return self.objclass([obj.outer(), obj.inner().add_cell(pos[0])])
        except:
            raise ValueError("Error removing cell %s from %s" % (pos, self.objclass))



# -*- coding: utf-8 -*-
r"""
Grid View Adapter for tableaux

**Grid View tableau operations:**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~TableauGridViewAdapter.compute_cells` | Compute tableau cells as a dictionary { coordinate pair : Integer }
    :meth:`~TableauGridViewAdapter.from_cells` | Create a new tableau from a cells dictionary
    :meth:`~TableauGridViewAdapter.addable_cells` | List addable cells
    :meth:`~TableauGridViewAdapter.add_cell` | Add a cell
    :meth:`~TableauGridViewAdapter.removable_cells` | List removable cells (Tableau)
    :meth:`~StandardTableauGridViewAdapter.removable_cells` | List removable cells (StandardTableau)
    :meth:`~TableauGridViewAdapter.remove_cell` | Remove a cell

AUTHORS ::

    Odile Bénassy, Nicolas Thiéry

"""
from sage.combinat.tableau import Tableau, StandardTableau, SemistandardTableau
from sage.rings.integer import Integer
from sage_widget_adapters.generic_grid_view_adapter import GridViewAdapter


class TableauGridViewAdapter(GridViewAdapter):
    r"""
    Grid view adapter for Young tableaux.

    ATTRIBUTES::
        * ``objclass`` -- Tableau
        * ``celltype`` -- Integer
        * ``cellzero`` -- Integer(0)
    """
    objclass = Tableau
    constructorname = 'Tableau'
    celltype = Integer # i.e. sage.rings.integer.Integer
    cellzero = Integer(0)

    @staticmethod
    def compute_cells(obj):
        r"""
        From a tableau,
        return a dictionary { coordinates pair : Integer }

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_widget_adapters.combinat.tableau_grid_view_adapter import TableauGridViewAdapter
            sage: t = Tableau([[1, 2, 5, 6], [3], [4]])
            sage: TableauGridViewAdapter.compute_cells(t)
            {(0, 0): 1, (0, 1): 2, (0, 2): 5, (0, 3): 6, (1, 0): 3, (2, 0): 4}
        """
        return {(i,j):obj[i][j] for (i,j) in obj.cells()}

    @classmethod
    def from_cells(cls, cells={}):
        r"""
        From a dictionary { coordinates pair : Integer }
        return a corresponding tableau

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_widget_adapters.combinat.tableau_grid_view_adapter import TableauGridViewAdapter
            sage: TableauGridViewAdapter.from_cells({(0, 0): 1, (0, 1): 2, (0, 2): 5, (0, 3): 6, (1, 0): 3, (2, 0): 4})
            [[1, 2, 5, 6], [3], [4]]
        """
        rows = []
        i = 0
        while i <= max(pos[0] for pos in cells):
            row = [cells[pos] for pos in cells if pos[0] == i]
            row.sort()
            rows.append(row)
            i += 1
        try:
            return cls.objclass(rows)
        except:
            raise TypeError(
                "This object is not compatible with this adapter (%s, for %s objects)" % (cls, cls.objclass))

    @staticmethod
    def addable_cells(obj, borders=False):
        r"""
        List object addable cells

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_widget_adapters.combinat.tableau_grid_view_adapter import TableauGridViewAdapter
            sage: t = Tableau([[1, 3, 4, 8, 12, 14, 15], [2, 7, 11, 13], [5, 9], [6, 10]])
            sage: TableauGridViewAdapter.addable_cells(t, True)
            ([(0, 7), (1, 4), (2, 2), (4, 0)], [(0, 7), (1, 4), (2, 2)], [(1, 4), (2, 2), (4, 0)])
        """
        addable_cells = obj.shape().outside_corners()
        if not borders:
            return addable_cells
        no_left_border = [pos for pos in addable_cells if pos[0] < len(obj)]
        no_top_border = []
        prev = None
        for pos in addable_cells:
            if prev and pos[0] and pos[1] < prev[1]:
               no_top_border.append(pos)
            prev = pos
        return addable_cells, no_left_border, no_top_border

    @staticmethod
    def removable_cells(obj):
        r"""
        List object removable cells

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_widget_adapters.combinat.tableau_grid_view_adapter import TableauGridViewAdapter
            sage: t = Tableau([[1, 2, 5, 6], [3, 7], [4]])
            sage: TableauGridViewAdapter.removable_cells(t)
            [(0, 3), (1, 1), (2, 0)]
        """
        return obj.corners()

    def add_cell(self, obj, pos, val, dirty={}):
        r"""
        Add cell

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_widget_adapters.combinat.tableau_grid_view_adapter import TableauGridViewAdapter
            sage: t = Tableau([[1, 2, 5, 6], [3, 7], [4]])
            sage: ta = TableauGridViewAdapter()
            sage: ta.add_cell(t, (3, 0), 8)
            [[1, 2, 5, 6], [3, 7], [4], [8]]
            sage: ta.add_cell(t, (1, 2), 8)
            [[1, 2, 5, 6], [3, 7, 8], [4]]
            sage: ta.add_cell(t, (2, 0), 9)
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(2, 0)' is not addable.
            sage: ta.add_cell(t, (3, 0), 8, dirty={(3, 0):8})
            [[1, 2, 5, 6], [3, 7], [4], [8]]
        """
        if not pos in self.addable_cells(obj):
            raise ValueError("Cell position '%s' is not addable." % str(pos))
        tl = self.make_dirty(obj.to_list(), dirty)
        if not pos in dirty:
            if pos[0] >= len(tl):
                tl.append([val])
            else:
                tl[pos[0]].append(val)
        return self._validate(tl)

    def remove_cell(self, obj, pos, dirty={}):
        r"""
        Remove cell

        TESTS ::

            sage: from sage.combinat.tableau import Tableau
            sage: from sage_widget_adapters.combinat.tableau_grid_view_adapter import TableauGridViewAdapter
            sage: t = Tableau([[1, 2, 5, 6], [3, 7], [4]])
            sage: ta = TableauGridViewAdapter()
            sage: ta.remove_cell(t, (1, 1))
            [[1, 2, 5, 6], [3], [4]]
            sage: ta.remove_cell(t, (2, 0))
            [[1, 2, 5, 6], [3, 7]]
            sage: ta.remove_cell(t, (2, 1))
            Traceback (most recent call last):
            ...
            ValueError: Cell position '(2, 1)' is not removable.
            sage: from sage.combinat.tableau import StandardTableau
            sage: from sage_widget_adapters.combinat.tableau_grid_view_adapter import StandardTableauGridViewAdapter
            sage: st = StandardTableau([[1, 2, 5, 6], [3, 7], [4]])
            sage: sta = StandardTableauGridViewAdapter()
            sage: sta.remove_cell(st, (1, 1))
            [[1, 2, 5, 6], [3], [4]]
            sage: sta.remove_cell(st, (2, 0))
            ValueError('the entries in a standard tableau must be in bijection with 1,2,...,n',)
        """
        if not pos in self.removable_cells(obj):
            raise ValueError("Cell position '%s' is not removable." % str(pos))
        tl = obj.to_list()
        tl = self.make_dirty(tl, dirty)
        tl[pos[0]].pop()
        if not tl[pos[0]]:
            tl.pop()
        tl = [r for r in tl if r] # do not keep any empty row before the test
        return self._validate(tl)

class SemistandardTableauGridViewAdapter(TableauGridViewAdapter):
    r"""
    Value will validate as semistandard tableau.
    """
    objclass = SemistandardTableau
    constructorname = 'SemistandardTableau'

class StandardTableauGridViewAdapter(SemistandardTableauGridViewAdapter):
    r"""
    Value will validate as standard tableau.
    """
    objclass = StandardTableau
    constructorname = 'StandardTableau'
